unit FrmRndSelector;

interface

uses

  Winapi.Windows, Winapi.Messages,

  System.SysUtils, System.Variants, System.Classes, System.IniFiles,

  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.ExtCtrls,

  knobs2013, Vcl.StdCtrls, Vcl.Buttons,

  KnobsUtils;


type

  TControlItem = record
    FControl             : TKnobsValuedControl;
    FControlRandomizable : Boolean;
    FCheckBox            : TCheckBox;
  end;

  PControlItem  = ^ TControlItem;
  TControlItems = TArray<TControlItem>;


  TFormRandomSelector = class(TForm)
    PanelTop: TPanel;
    BitBtnCancel: TBitBtn;
    BitBtnAccept: TBitBtn;
    CheckBoxModuleRandomizable: TCheckBox;
    BitBtnSelectAll: TBitBtn;
    BitBtn1: TBitBtn;
    ScrollBoxMain: TScrollBox;
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure BitBtnCancelClick(Sender: TObject);
    procedure BitBtnAcceptClick(Sender: TObject);
    procedure BitBtnSelectAllClick(Sender: TObject);
    procedure BitBtn1Click(Sender: TObject);
  private
    FModule             : TKnobsCustomModule;
    FModuleRandomizable : Boolean;
    FControlItems       : TControlItems;
  private
    procedure   SetModuleRandomizable( aValue: Boolean);
  private
    procedure   Accept;
    procedure   Cancel;
    procedure   SelectAll;
    procedure   SelectNone;
    procedure   ClearControlItems;
    procedure   AddControlItem( const aControl: TKnobsValuedControl);
    procedure   SetupControls;
  private
    property    ModuleRandomizable: Boolean read FModuleRandomizable write SetModuleRandomizable;
  public
    procedure   Execute( const aModule: TKnobsCustomModule);
    procedure   LoadIni( const aSection: string; anIniFile: TCustomIniFile);
    procedure   SaveIni( const aSection: string; anIniFile: TCustomIniFile);
  end;

var


  FormRandomSelector: TFormRandomSelector;

implementation


// User area

//  private

    procedure   TFormRandomSelector.SetModuleRandomizable( aValue: Boolean);
    begin
      if   aValue <> FModuleRandomizable
      then begin
        FModuleRandomizable                := aValue;
        CheckBoxModuleRandomizable.Checked := aValue;
      end;
    end;


//  private

    procedure   TFormRandomSelector.Accept;
    var
      i : Integer;
    begin
      if   Assigned( FModule)
      then begin
        FModule.AllowRandomization := ModuleRandomizable;

        for i := 0 to Length( FControlItems) - 1
        do FControlItems[ i].FControl.AllowRandomization := FControlItems[ i].FCheckBox.Checked;
      end;

      ModalResult := mrOk;
    end;


    procedure   TFormRandomSelector.Cancel;
    begin
      ModalResult := mrCancel;
    end;


    procedure   TFormRandomSelector.SelectAll;
    var
      i : Integer;
    begin
      ModuleRandomizable := True;

      for i := 0 to Length( FControlItems) - 1
      do FControlItems[ i].FCheckBox.Checked := True;
    end;


    procedure   TFormRandomSelector.SelectNone;
    var
      i : Integer;
    begin
      ModuleRandomizable := False;

      for i := 0 to Length( FControlItems) - 1
      do FControlItems[ i].FCheckBox.Checked := False;
    end;


    procedure   TFormRandomSelector.ClearControlItems;
    var
      i : Integer;
    begin
      for i := 0 to Length( FControlItems) - 1
      do FreeAndNil( FControlItems[ i].FCheckBox);

      SetLength( FControlItems, 0);
    end;


    procedure   TFormRandomSelector.AddControlItem( const aControl: TKnobsValuedControl);
    var
      L : Integer;
      C : TCheckBox;
    begin
      if   Assigned( aControl)
      then begin
        L := Length( FControlItems);
        SetLength( FControlItems, L + 1);
        FControlItems[ L].FControl             := aControl;
        FControlItems[ L].FControlRandomizable := aControl.AllowRandomization;
        C                                      := TCheckBox.Create( Self);
        C.Checked                              := aControl.AllowRandomization;
        C.Caption                              := Format( '%s  ( %s)', [ aControl.ShortName, aControl.FriendlyName], AppLocale);
        C.Tag                                  := NativeInt( @ FControlItems[ L]);
        C.Parent                               := ScrollBoxMain;
        C.Top                                  := L;
        C.TabOrder                             := L;
        C.Align                                := alTop;
        FControlItems[ L].FCheckBox            := C;
      end;
    end;


    procedure   TFormRandomSelector.SetupControls;
    var
      i : Integer;
    begin
      if   Assigned( FModule)
      then begin
        CheckboxModuleRandomizable.Caption := Format( 'Module (%s) randomizable', [ FModule.Title], AppLocale);
        ModuleRandomizable := FModule.AllowRandomization;

        for i := 0 to FModule.ControlCount - 1
        do begin
          if   FModule.Controls[ i] is TKnobsValuedControl
          and  TKnobsValuedControl( FModule.Controls[ i]).AllowAutomation
          then AddControlItem( TKnobsValuedControl( FModule.Controls[ i]));
        end;
      end;
    end;


//  public

    procedure   TFormRandomSelector.Execute( const aModule: TKnobsCustomModule);
    begin
      ClearControlItems;
      FModule := aModule;
      SetupControls;
      ShowModal;
    end;


    procedure   TFormRandomSelector.LoadIni( const aSection: string; anIniFile: TCustomIniFile);
    var
      L, T, W, H    : Integer;
    begin
      if   Assigned( anIniFile)
      and  ( aSection <> '')
      then begin
        L := anIniFile.ReadInteger( aSection, 'Left'  , Left  );
        T := anIniFile.ReadInteger( aSection, 'Top'   , Top   );
        W := anIniFile.ReadInteger( aSection, 'Width' , Width );
        H := anIniFile.ReadInteger( aSection, 'Height', Height);
        SetBounds( L, T, W, H);
      end;
    end;


    procedure   TFormRandomSelector.SaveIni( const aSection: string; anIniFile: TCustomIniFile);
    begin
      if   Assigned( anIniFile)
      and  ( aSection <> '')
      then begin
        anIniFile.EraseSection( aSection);
        anIniFile.WriteInteger( aSection, 'Left'  , Left  );
        anIniFile.WriteInteger( aSection, 'Top'   , Top   );
        anIniFile.WriteInteger( aSection, 'Width' , Width );
        anIniFile.WriteInteger( aSection, 'Height', Height);
      end;
    end;


// Delphi are

{$R *.dfm}

procedure TFormRandomSelector.BitBtn1Click(Sender: TObject);
begin
  SelectNone;
end;

procedure TFormRandomSelector.BitBtnAcceptClick(Sender: TObject);
begin
  Accept;
end;

procedure TFormRandomSelector.BitBtnCancelClick(Sender: TObject);
begin
  Cancel;
end;

procedure TFormRandomSelector.BitBtnSelectAllClick(Sender: TObject);
begin
  SelectAll;
end;

procedure TFormRandomSelector.FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if   Word( Key) = VK_RETURN
  then begin
    Key := 0;
    Accept;
  end
  else if Word( Key) = VK_ESCAPE
  then begin
    Key := 0;
    Cancel;
  end;
end;

end.
