unit FrmSearchResults;

interface

uses

  WinApi.Windows,

  System.SysUtils, System.Classes, System.IniFiles,

  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.ExtCtrls, Vcl.Buttons, Vcl.Grids,

  KnobsUtils;


type

  TSearchResult = record
    ModuleType : Integer;
    ModuleName : string;
    ModuleText : string;
  end;

  TSearchResults = array of TSearchResult;

   TCloseAction = (
     srcaClose,
     srcaShowHelp,
     srcaShowModule
   );

   TOnSearchClose = procedure( const aSender: TObject; aModueType: Integer; anAction: TCloseAction) of object;

  TFormSearchResults = class(TForm)
    PanelTop: TPanel;
    BitBtnClose: TBitBtn;
    BitBtnShowHelp: TBitBtn;
    BitBtnShowInSelector: TBitBtn;
    StringGridSearchResults: TStringGrid;
    LabelModuleType: TLabel;
    procedure BitBtnCloseClick(Sender: TObject);
    procedure BitBtnShowHelpClick(Sender: TObject);
    procedure BitBtnShowInSelectorClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure StringGridSearchResultsClick(Sender: TObject);
    procedure StringGridSearchResultsDblClick(Sender: TObject);
    procedure StringGridSearchResultsKeyPress(Sender: TObject; var Key: Char);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
  private
    FOnSearchClose : TOnSearchClose;
    FModuleType    : Integer;
    FCanClose      : Boolean;
  private
    procedure   SetModuleType( aValue: Integer);
    procedure   HandleCloseCallback( anAction: TCloseAction);
    procedure   CloseSearch        ( anAction: TCloseAction);
    procedure   CloseForm;
    procedure   ShowHelp;
    procedure   ShowModule;
    procedure   SetupForm;
    procedure   HandleFormResize;
    procedure   FixRowSelection;
  private
    property    ModuleType : Integer read FModuleType write SetModuleType;
  public
    procedure   ShowSearchResults( const aList: TSearchResults);
    procedure   LoadIni( const aSection: string; anIniFile: TCustomIniFile);
    procedure   SaveIni( const aSection: string; anIniFile: TCustomIniFile);
  public
    property    OnSearchClose : TOnSearchClose read FOnSearchClose write FOnSearchClose;
  end;

var

  FormSearchResults: TFormSearchResults;


implementation


// User area

//  private

    procedure   TFormSearchResults.SetModuleType( aValue: Integer);
    begin
      if   aValue <> FModuleType
      then begin
        FModuleType                  := aValue;
        BitBtnShowHelp      .Enabled := ModuleType >= 0;
        BitBtnShowInSelector.Enabled := ModuleType >= 0;
      end;
    end;


    procedure   TFormSearchResults.HandleCloseCallback( anAction: TCloseAction);
    begin
      if   Assigned( FOnSearchClose)
      and  ( ModuleType >= 0)
      then FOnSearchClose( Self, FModuleType, anAction);
    end;


    procedure   TFormSearchResults.CloseSearch( anAction: TCloseAction);
    begin
      HandleCloseCallback( anAction);
      Close;
    end;


    procedure   TFormSearchResults.CloseForm;
    begin
      CloseSearch( srcaClose);
    end;


    procedure   TFormSearchResults.ShowHelp;
    begin
      CloseSearch( srcaShowHelp);
    end;


    procedure   TFormSearchResults.ShowModule;
    begin
      if   ModuleType >= 0
      then CloseSearch( srcaShowModule)
      else CloseSearch( srcaClose     );
    end;


    procedure   TFormSearchResults.SetupForm;
    begin
      StringGridSearchResults.ColCount      := 3;
      StringGridSearchResults.FixedCols     := 0;
      StringGridSearchResults.RowCount      := 2;
      StringGridSearchResults.FixedRows     := 1;
      StringGridSearchResults.Cells[ 0, 0]  := 'Type';
      StringGridSearchResults.Cells[ 1, 0]  := 'Name';
      StringGridSearchResults.Cells[ 2, 0]  := 'Text';
      StringGridSearchResults.ColWidths[ 0] :=  64;
      StringGridSearchResults.ColWidths[ 1] := 136;
      HandleFormResize;
      FCanClose := False;
    end;


    procedure   TFormSearchResults.HandleFormResize;
    var
      FixedWidth : Integer;
    begin
      FixedWidth := StringGridSearchResults.ColWidths[ 0] + StringGridSearchResults.ColWidths[ 1] + 4;
      StringGridSearchResults.ColWidths[ 2] := StringGridSearchResults.Width - FixedWidth;
    end;


    procedure   TFormSearchResults.FixRowSelection;
    begin
      if   StringGridSearchResults.Row > 0
      then begin
        ModuleType := StrToIntDef( StringGridSearchResults.Cells[ 0, StringGridSearchResults.Row], -1);
        LabelModuleType.Caption := Format( '%d %s', [ ModuleType, StringGridSearchResults.Cells[ 1, StringGridSearchResults.Row]], AppLocale);
      end
      else LabelModuleType.Caption := 'nothing selected';
    end;


//  public

    procedure   TFormSearchResults.ShowSearchResults( const aList: TSearchResults);
    var
      i : Integer;
      L : Integer;
    begin
      ModuleType := -1;
      L          := Length( aList);

      if   L = 0
      then begin
        StringGridSearchResults.RowCount := 2;
        StringGridSearchResults.Cells[ 0, 1] := '';
        StringGridSearchResults.Cells[ 1, 1] := '';
        StringGridSearchResults.Cells[ 2, 1] := '';
      end
      else StringGridSearchResults.RowCount  := L + 1;

      StringGridSearchResults.FixedRows := 1;

      for i := 0 to L - 1
      do begin
        StringGridSearchResults.Cells[ 0, i + 1] := IntToStr( aList[ i].ModuleType);
        StringGridSearchResults.Cells[ 1, i + 1] :=           aList[ i].ModuleName ;
        StringGridSearchResults.Cells[ 2, i + 1] :=           aList[ i].ModuleText ;
      end;

      FixRowSelection;
    end;


    procedure   TFormSearchResults.LoadIni( const aSection: string; anIniFile: TCustomIniFile);
    var
      L, T, W, H    : Integer;
    begin
      if   Assigned( anIniFile)
      and  ( aSection <> '')
      then begin
        L := anIniFile.ReadInteger( aSection, 'Left'  , Left  );
        T := anIniFile.ReadInteger( aSection, 'Top'   , Top   );
        W := anIniFile.ReadInteger( aSection, 'Width' , Width );
        H := anIniFile.ReadInteger( aSection, 'Height', Height);
        SetBounds( L, T, W, H);
      end;
    end;


    procedure   TFormSearchResults.SaveIni( const aSection: string; anIniFile: TCustomIniFile);
    begin
      if   Assigned( anIniFile)
      and  ( aSection <> '')
      then begin
        anIniFile.EraseSection( aSection);
        anIniFile.WriteInteger( aSection, 'Left'  , Left  );
        anIniFile.WriteInteger( aSection, 'Top'   , Top   );
        anIniFile.WriteInteger( aSection, 'Width' , Width );
        anIniFile.WriteInteger( aSection, 'Height', Height);
      end;
    end;


// Delphi area

{$R *.dfm}

procedure TFormSearchResults.StringGridSearchResultsClick(Sender: TObject);
begin
  FixRowSelection;
end;

procedure TFormSearchResults.StringGridSearchResultsDblClick(Sender: TObject);
begin
  FixRowSelection;
  ShowModule;
end;

procedure TFormSearchResults.StringGridSearchResultsKeyPress(Sender: TObject; var Key: Char);
begin
  FixRowSelection;
end;

procedure TFormSearchResults.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  CanClose := True;
  HandleCloseCallback( srcaClose);
end;

procedure TFormSearchResults.FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if      Word( Key) = VK_RETURN then begin Key := 0; ShowModule; end
  else if Word( KEY) = VK_ESCAPE then begin Key := 0; CloseForm;  end
  else if Word( KEY) = VK_F1     then begin Key := 0; ShowHelp;   end;
end;

procedure TFormSearchResults.FormResize(Sender: TObject);
begin
  HandleFormResize;
end;

procedure TFormSearchResults.FormShow(Sender: TObject);
begin
  SetupForm;
end;

procedure TFormSearchResults.BitBtnShowHelpClick(Sender: TObject);
begin
  ShowHelp;
end;

procedure TFormSearchResults.BitBtnShowInSelectorClick(Sender: TObject);
begin
  ShowModule;
end;

procedure TFormSearchResults.BitBtnCloseClick(Sender: TObject);
begin
  CloseForm;
end;

end.
