unit Logger;

interface

  procedure WriteLog( const aMsg: string);
  procedure WriteLogFmt( const aFmt: string; const anArgs: array of const);


implementation


uses

  WinApi.Windows,

  System.Classes, System.IOUtils, System.SysUtils, System.SyncObjs, System.Generics.Collections,

  Vcl.Forms,

  KnobsUtils;


type

  TLogger = class
  private
    FLock     : TCriticalSection;
    FFileName : string;
  public
    procedure   AddEvent( const aMsg: string);
    constructor Create( const aFileName: string);
    destructor  Destroy;                                                                                       override;
  end;


{ ========
  TLogger = class( TThread)
  private
    FLock     : TCriticalSection;
    FData     : TQueue<string>;
    FFileName : string;
  public
}

    procedure   TLogger.AddEvent( const aMsg: string);
    var
      aFile : TextFile;
      S     : string;
    begin
      FLock.Enter;

      try
        AssignFile( aFile, FFileName);

        if   FileExists( FFileName)
        then Append ( aFile)
        else Rewrite( aFile);

        try
          S :=  FormatDateTime( 'yyyy-mm-dd hh:nn:ss.zzz', Now) + ' - ' + aMsg;
          WriteLn( aFile, S);
        finally
          CloseFile( aFile);
        end;
      finally
        FLock.Leave;
      end;
    end;

    constructor TLogger.Create( const aFileName: string);
    begin
      FLock     := TCriticalSection.Create;
      FFileName := aFileName;

      if   FileExists( FFileName)
      then DeleteFile( FFileName);
    end;


    destructor  TLogger.Destroy; // override;
    begin
      FLock.Enter;
      FLock.Leave;
      FreeAndNil( FLock);
    end;



var

  FLogger : TLogger;


  procedure WriteLog( const aMsg: string);
  begin
    FLogger.AddEvent( aMsg);
  end;


  procedure WriteLogFmt( const aFmt: string; const anArgs: array of const);
  begin
    WriteLog( Format( aFmt, anArgs, AppLocale));
  end;


  procedure InitLogger;
  begin
    FLogger := TLogger.Create( IncludeTrailingPathDelimiter( ApplicationPath) + 'vren.txt');
    WriteLog( '** Log started');
  end;


  procedure DoneLogger;
  begin
    WriteLog( '** Log done');
    FreeAndNil( FLogger);
  end;


initialization

  InitLogger;

finalization

  DoneLogger;

end.
