unit VstDspWrapper;

interface

uses

  module_defs,

  KnobsUtils, KnobsConversions, midi_defs;

type


  TVrenSynth = class
  private
    FSynthesizer : TSynthesizer;
    FTickInData  : TSignalArray;  // Audio samples passed into the synth
    FTickOutData : TSignalArray;
    FLockCount   : Integer;
  private
    function    GetLocked: Boolean;
    procedure   SetLocked( aValue: Boolean);
  public
    constructor Create( aSampleRate: Single; aPolyphony: Integer);
    destructor  Destroy;                                                                                       override;
    function    ChangeSynthesizer( const aSynthesizer: TSynthesizer): TSynthesizer;
    procedure   UpdateParameter( anId: Integer; aValue: Single);
    procedure   AcceptMidi ( const aMsg: TMidiMessage);
    procedure   AcceptAudio( const aData: TSignalArray);
  public
    property    Locked       : Boolean      read GetLocked       write SetLocked;
  end;


implementation

uses

  Logger;


{ ========
  TVrenSynth = class
  private
    FSynthesizer    : TSynthesizer;
    FTickInData     : TSignalArray;  // Audio samples passed into the synth
    FTickOutData    : TSignalArray;
    FLockCount      : Integer;
    FInMaxima       : TSignalArray;
    FInMinima       : TSignalArray;
    FOutMaxima      : TSignalArray;
    FOutMinima      : TSignalArray;
    FFixedInMaxima  : TSignalArray;
    FFixedInMinima  : TSignalArray;
    FFixedOutMaxima : TSignalArray;
    FFixedOutMinima : TSignalArray;
    FFrameWrapped   : Integer;
    FFrameCount     : Integer;
    FFrameCounter   : Integer;
  public
    property    Locked       : Boolean      read GetLocked       write SetLocked;
    property    FrameWrapped : Boolean      read GetFrameWrapped;
    property    InMaxima     : TSignalArray read FFixedInMaxima;
    property    InMinima     : TSignalArray read FFixedInMinima;
    property    OutMaxima    : TSignalArray read FFixedOutMaxima;
    property    OutMinima    : TSignalArray read FFixedOutMinima;
  private
}

    function    TVrenSynth.GetLocked: Boolean;
    begin
      Result := FLockCount > 0;
    end;


    procedure   TVrenSynth.SetLocked( aValue: Boolean);
    begin
      if   aValue
      then Inc( FLockCount)
      else DecToZero( FLockCount);
    end;


//  public

    constructor TVrenSynth.Create( aSampleRate: Single; aPolyphony: Integer);
    begin
      inherited Create;
      FSynthesizer := TSynthesizer.Create( nil, 'vren', 10, aPolyphony);
      SetLength( FTickInData , MAX_CHANNELS);
      SetLength( FTickOutData, MAX_CHANNELS);
    end;


    destructor  TVrenSynth.Destroy; // override;
    var
      OldSynth: TSynthesizer;
    begin
      OldSynth := ChangeSynthesizer( nil);
      OldSynth.DisposeOf;
      inherited;
    end;


    function    TVrenSynth.ChangeSynthesizer( const aSynthesizer: TSynthesizer): TSynthesizer;
    begin
      Result := AtomicExchange( Pointer( FSynthesizer), Pointer( aSynthesizer));
    end;


    procedure   TVrenSynth.UpdateParameter( anId: Integer; aValue: Single);
    begin
    end;

    procedure   TVrenSynth.AcceptMidi( const aMsg: TMidiMessage);
    begin
      if   Assigned( FSynthesizer)
      and  not Locked
      then FSynthesizer.AcceptMidi( aMsg);
    end;


    procedure   TVrenSynth.AcceptAudio( const aData: TSignalArray);
    var
      i : Integer;
    begin
          if   Assigned( FSynthesizer)
      then begin
        if   Locked
        then begin
          for i := 0 to Length( aData) - 1
          do begin
            FTickInData [ i] := aData[ i];
            FTickOutData[ i] := 0.0;
            aData       [ i] := 0.0;
          end;
        end
        else begin
          for i := 0 to Length( aData) - 1
          do FTickInData[ i] := aData[ i];

          FSynthesizer.TickWithData( aData);

          for i := 0 to Length( aData) - 1
          do begin
            aData[ i]        := FSynthesizer.PolyOutputs[ i];
            FTickOutData[ i] := aData[ i];
          end;
        end;
      end;

      //  for i := 0 to Length( aData) - 1
      //  do aData[ i] := ( 0.1 * Random) * ( 1.0 - 2.0 * Random);

    end;


end.

