unit VstWrapper;

interface

uses


  Forms, Classes,

  Vst3Base, UVSTInstrument, UVSTBase, VstDspWrapper,

  KnobsUtils, KnobsConversions, midi_defs,

  FrmDllMain;


type

  TVstWrapper = class( TVSTInstrument)
  private
    FSynthesizer    : TVrenSynth;
    FSynthData      : TSignalArray;
    FMidiMessage    : TMidiMessage;
  protected
    procedure   DoKeyDown( const aSender: TObject; aKey: Char16; aKeyCode, aModifiers: Int16);
    procedure   DoKeyUp  ( const aSender: TObject; aKey: Char16; aKeyCode, aModifiers: Int16);
    procedure   OnInitialize;                                                                                  override;
    procedure   Process32( aSamples, aChannels: Integer; anInput, anOutput: PPSingle);                         override;
    procedure   OnMidiEvent( aByte0, aByte1, aByte2: Integer);                                                 override;
    procedure   UpdateProcessorParameter( anId: Integer; aValue: Double);                                      override;
    procedure   UpdateEditorParameter( anId: Integer; aValue:Double);                                          override;
  protected
    procedure   OnEditOpen;                                                                                    override;
  public
    procedure   LockSynth;
    procedure   UnlockSynth;
  end;

  function GetVSTInstrumentInfo: TVSTInstrumentInfo;


implementation


{ TmyVST }

uses

  WinApi.Windows, WinApi.Messages,

  System.SysUtils,

  UCConsts, Logger, KeyTranslations;


{$POINTERMATH ON}
{$define DebugLog}

{ ========
  TMyVSTPlugin = class( TVSTInstrument)
  private
    FSynthesizer    : TVrenSynth;
    FSynthData      : TSignalArray;
    FInMaxima       : TSignalArray;
    FInMinima       : TSignalArray;
    FOutMaxima      : TSignalArray;
    FOutMinima      : TSignalArray;
    FFixedInMaxima  : TSignalArray;
    FFixedInMinima  : TSignalArray;
    FFixedOutMaxima : TSignalArray;
    FFixedOutMinima : TSignalArray;
    FMidiMessage    : TMidiMessage;
    FFrameWrapped   : Integer;
    FFrameCount     : Integer;
    FFrameCounter   : Integer;
  public
    property    FrameWrapped : Boolean      read GetFrameWrapped;
    property    InMaxima     : TSignalArray read FFixedInMaxima;
    property    InMinima     : TSignalArray read FFixedInMinima;
    property    OutMaxima    : TSignalArray read FFixedOutMaxima;
    property    OutMinima    : TSignalArray read FFixedOutMinima;
  private
}

//  protected

    procedure   TVstWrapper.DoKeyDown( const aSender: TObject; aKey: Char16; aKeyCode, aModifiers: Int16);
//    var
//      aMsg : TWmKeyDown;
    begin
//      if   Assigned( FormDllMain)
//      and  VstValidKeyCode( aKeyCode)
//      then begin
//        // aMsg := VstKeyDownToWinKeyDown( aKey, aKeyCode, aModifiers);
//        // PostMessage( FormDllMain.Handle, aMsg);
//        FormDllMain.AcceptKeyDown( aSender, aKey, aKeyCode, aModifiers)
//      end
//      else
      WriteLogFmt( 'TVstWrapper.DoKeyDown unhandled key %s', [ VstKeyDataToStr( aKey, aKeyCode, aModifiers)]);
    end;


    procedure   TVstWrapper.DoKeyUp( const aSender: TObject; aKey: Char16; aKeyCode, aModifiers: Int16);
    begin
//      if   Assigned( FormDllMain)
//      and  VstValidKeyCode( aKeyCode)
//      then begin
//        FormDllMain.AcceptKeyUp( aSender, aKey, aKeyCode, aModifiers)
//      end
//      else
      WriteLogFmt( 'TVstWrapper.DoKeyUp unhandled key %s', [ VstKeyDataToStr( aKey, aKeyCode, aModifiers)]);
    end;


    procedure   TVstWrapper.OnInitialize;
    begin
      SetLength( FSynthData, MAX_CHANNELS);
      // SetLength( FMidiMessage.Data, 2); .. no longer needed

      // This will not work .. as FormDllMain never gets instantiated
      // FSynthesizer := TVrenSynth.Create( FormDllMain, 48000);
      FSynthesizer := TVrenSynth.Create( 48000, 1);   // todo: must set a proper polyphony here ...
      OnKeyDown    := DoKeyDown;
      OnKeyUp      := DoKeyUp;

    //  AddParameter( ID_CUTOFF    ,'Cutoff'     , 'Cutoff'   , 'Hz', 20,20000, 10000);
    //  AddParameter( ID_RESONANCE ,'Resonance'  , 'Resonance', ''  ,  0,    1,     0);
    //  AddParameter( ID_PULSEWIDTH,'Pulse Width', 'PWM'      , '%' ,  0,  100,    50);
    //  AddProgram('Program 1');
    //  AddProgram('Program 2');
    //  AddProgram('Program 3');
    end;


    procedure   TVstWrapper.Process32( aSamples, aChannels: Integer; anInput, anOutput: PPSingle);
    var
      Frame   : Integer;
      Channel : Integer;
    begin
      if   Assigned( FSynthesizer)
      and  ( Length( FSynthData) = 8)
      then begin
        for Frame := 0 to aSamples - 1
        do begin
          for Channel := 0 to MAX_CHANNELS - 1
          do begin
            if   Channel < aChannels
            then FSynthData[ Channel] := anInput[ Channel, Frame]
            else FSynthData[ Channel] := 0.0;
          end;

          FSynthesizer.AcceptAudio( FSynthData);

          for Channel := 0 to MAX_CHANNELS - 1
          do  begin
            if   Channel < aChannels
            then anOutput[ Channel][ Frame] := FSynthData[ Channel];
          end;
        end;
      end;
    end;


    procedure   TVstWrapper.OnMidiEvent( aByte0, aByte1, aByte2: Integer);
    begin
      // Note: OnMidiEvent can be called from NON-UI thread. (in VST3) Update UI on own risk
      //
      // Note: for parameter based changes, you only have to update the processor
      // we COULD make an extra method in the fw OnMidiEventEditor, but that should be discussed
      // this is perhaps one of the reasons Steinberg chose to 'depricate' midi stuff

      // WriteLog( 'TMyVSTPlugin.OnMidiEvent:' + aByte0.ToString + ' ' + aByte1.ToString + ' ' + aByte2.ToString);

      if   Assigned( FSynthesizer)
      then begin
        FMidiMessage.STime   := 0;
        FMidiMessage.Command := aByte0 and $f0;
        FMidiMessage.Channel := aByte0 and $0f;
        FMidiMessage.Data1   := aByte1;
        FMidiMessage.Data2   := aByte2;
        FSynthesizer.AcceptMidi( FMidiMessage);
      end;
    end;


    procedure   TVstWrapper.UpdateProcessorParameter( anId:Integer; aValue: Double);
    begin
      FSynthesizer.UpdateParameter( anId, aValue);
    end;


    procedure TVstWrapper.UpdateEditorParameter( anId: Integer; aValue: Double);
    begin
    end;


//  protected

    procedure   TVstWrapper.OnEditOpen;
    begin
      ResendParameters;
    end;


//  public

    procedure   TVstWrapper.LockSynth;
    begin
      FSynthesizer.Locked := True;
    end;


    procedure   TVstWrapper.UnlockSynth;
    begin
      FSynthesizer.Locked := False;
    end;


const

  VST3_ID: TGUID = '{9137E7DE-95BD-490A-BEDB-A564685982E0}';

    function GetVSTInstrumentInfo: TVSTInstrumentInfo;
    begin
      with Result
      do begin
        with PluginDef
        do begin
          Vst3Id        := VST3_ID;
          InstrClass    := TVstWrapper;
          Name          := 'vren';
          ExternalClass := TFormDllMain;
          IsSynth       := True;
          Vst2Id        := 'VREN';
        end;

        with FactoryDef
        do begin
          Vendor := 'Blue Hell';
          URL    := 'bluehell.nl';
          EMail  := 'jan/@/mail.bluehell.nl';
        end;
      end;
    end;

end.

