 unit wave_tools;

{

   COPYRIGHT 2013 .. 2022 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    https://bluehell.nl/
    jan2@t2mail_dot_bluehell_dot_nl


  Parts were copied and modified from software found on the webs ... it's blurp was:

  +---------------------------------------------------------------------------------------------+
  | **** Note to original developer : NEVER EVER FOKKING PUT FUNCTIONAL CODE IN assertions **** |
  | (as assertions can be optimized out for the release, and the code will no longer work then) |
  | (and it took me half a day to find why the released code was not working                  ) |
  *---------------------------------------------------------------------------------------------+

  (*
    RiffRW is a RIFF file reader and writer, it uses Delphi memory streams to
    allow the user to add/change/delete data from each component of the RIFF file.
    Users can create new chunks, remove unwanted chunks and edit the content of
    existing chunks with great ease since work is based on the well known Delphi
    memory stream class.

    Note: this unit is released as freeware. In other words, you are free
          to use this unit in your own applications, however I retain all
          copyright to the code. LS

    Liran Shahar, Com-N-Sense Ltd (C)Copyright 2002, all rights reserved
  *)

  I didn't care about the interface stuff used in that one, so skipped all that
  and then added all the difficult stuff ...
}


interface

uses

  System.Classes, System.SysUtils, System.Math,

  KnobsUtils;


const
                                              // --- ACID ---
  FLACID_ONE_SHOT      = $01;                 // on: one shot        , off: loop
  FLACID_ROOT_NOTE_SET = $02;                 // on: root note is set, off: no root
  FLACID_STRETCH       = $04;                 // on: stretch is on   , off: stretch is off
  FLACID_DISK_BASED    = $08;                 // on: disk based      , off: ram based
  FLACID_ACIDIZER      = $10;                 // ?? acidizer turns it on ... or something


type

  TCueType = (
    ctUnknown,
    ctLabel  ,
    ctRegion ,
    ctLoop
  );

  TFourCC   = packed array[ 0 .. 3] of AnsiChar;
  TExtArray = packed array[ 0 .. 9] of Byte;


  TFormatInfo = packed record                 // ---  WAV --- used for AIF too, but then calculated from 'COMM' section
  public
    WaveFormat   : Word;                      //  0 | Compression code ...
    ChannelCount : Word;                      //  2 | Number of channels
    SampleRate   : Cardinal;                  //  4 | Sample rate
    StreamSpeed  : Cardinal;                  //  8 | Average bytes per second
    BlockAlign   : Word;                      // 12 | Block align
    BitCount     : Word;                      // 14 | Bits per sample
    ExtraCount   : Word;                      // 16 | Number of extra bytes
 // [Extra]      : array[ 1 .. ExteraCount] of Byte; // 18 | optional extra data
  end;                                        // 18 + ExtraCount


  TAcidInfo = packed record                   // ---  WAV ---
  public
    FileType   : Cardinal;                    //  0 | Combination of FL_ACID_XXX flags
    RootNote   : Word;                        //  4 | FLACID_ACIDIZER set : [C, C# .. , B] -> [$30 - $3B]
                                              //    | FLACID_ACIDIZER clr : [C, C# .. , B] -> [$3C - $47]
    UnSpec_1   : Word;                        //  6 | ?? always $8000 .. hmm .. no .. seen $0000 too
    UnSpec_2   : Single;                      //  8 | ?? seems to be always 0
    NumBeats   : Cardinal;                    // 12 | Number of beats
    MeterNumer : Word;                        // 16 | 3 for 3 / 4 measure
    MeterDenom : word;                        // 18 | 4 for 3 / 4 measure
    Tempo      : Single;                      // 20 | in BPM I guess ...
  end;                                        // 24 | and 24 indeed is observed size ...


  TLoopInfo = packed record                   // ---  common ---
  public
    Id         : Cardinal;                    // Loop ID - may be mapped to an ASCII name trough an "cue " chunk
                                              //         - name data should be in a "LIST" "adtl" chunk
    LoopType   : Cardinal;                    // 0 - forward, 1 - alternating fwd bwd, 2 - backward, 3-31: reserved, 32.. sampler specific
    LoopStart  : Cardinal;                    // First sample played for the loop in "data" chunk
    LoopEnd    : Cardinal;                    // Last  sample played for the loop in "data" chunk
    Fraction   : Cardinal;                    // fine tuning for loop fractional areas .. $80000000 -> 1 / 2 sample
    PlayCount  : Cardinal;                    // Repeat count for the loop
  end;


  TSampleInfo = packed record                 // ---  WAV ---
  public
    Manufacturer      : Cardinal;             // MMA manufacturere
    Product           : Cardinal;             // Manufacturer's product code
    SamplePeriod      : Cardinal;             // Period of one sample in 1/nSamplesPerSec in ns; 44k1 -> 22675
    MIDIUnityNote     : Cardinal;             // Midi note playing sample at original pitch
    MIDIPitchFraction : Cardinal;             // Semitone fraction $80000000 -> 50 cents
    SMPTEFormat       : Cardinal;             // 0 - none, 24 - 24 fps, 25 - 25 fps, 29 - 30 fps drop, 30 - 30 fps
    SMPTEOffset       : Cardinal;             // time offset, not used when Format = 0
    SampleLoops       : Cardinal;             // Number of following TLoopInfoStructs (in [LoopData])
    SamplerDataSize   : Cardinal;             // Size in bytes of sampler data, after [LoopData]
 // [LoopData]        : array[ 1 .. SampleLoops] of TLoopInfo
 // [SamplerData]     : array of Byte
  end;


  TCueInfo = packed record                    // ---  common ---
  public
    Id           : Cardinal;
    Position     : Cardinal;
    Chunk        : TFourCC;
    ChunkStart   : Cardinal;
    BlockStart   : Cardinal;
    SampleOffset : Cardinal;
  end;


  TCueName = record                           // ---  common ---
  public
    Id           : Cardinal;
    Name         : string;
    CueType      : TCueType;
    SampleLength : Cardinal;
  public
    function    CueTypeName: string;
  end;


  TCommInfo = packed record                   // ---  AIF ---
  public
    numChannels     : SmallInt;               //  0 | Audio channels
    numSampleFrames : Cardinal;               //  2 | Samples / channel
    SampleSize      : SmallInt;               //  6 | Bits / Sample
    SampleRate      : TExtArray;              //  8 | SampleFrams / sec - documented as extended .. but seems to not be IEEE ...
  public                                      // 18
    procedure   DoByteSwaps;
    function    GetSampleRate: Cardinal;      // Sample rate converted from extended to cardinal
  end;


  TCommInfoExt = packed record                // --- ? ---
  public
    numChannels     : SmallInt;               //  0 | Audio channels
    numSampleFrames : Cardinal;               //  2 | Samples / channel
    SampleSize      : SmallInt;               //  6 | Bits / Sample
    SampleRate      : TExtArray;              //  8 | SampleFrams / sec - documented as extended .. but seems to not be IEEE ...
    CompressionType : TFourCC;                // 18 | The compression type used
    StrLength       : Byte;                   // 22 The amount of string data that will follo
 // StringData      : array[ StrLength] of AnsiChar
  public
    procedure   DoByteSwaps;
    function    GetSampleRate: Cardinal;      // Sample rate converted from extended to cardinal
  end;


  TSF2Info = record                           // --- SF2 --- from <LIST><INFO>
  public
    Valid           : Boolean;                // True when data was acceptable
    VersionMajor    : Word;                   // ifil sub chunk
    VersionMinor    : Word;                   // ifil sub chunk
    SoundEngine     : string;                 // isng sub chunk
    BankName        : string;                 // INAM sub chunk
    CreationDate    : string;                 // ICRD sub chunk - verbatim, whatever the format that is in the file
    Engineers       : string;                 // IENG sub chunk - engineers - verbatim
    ProductId       : string;                 // IPRD sub chunk - the product the font was made for, verbatim
    Copyright       : string;                 // ICOP sub chunk - copyright, verbatim
    Comment         : string;                 // ICMT sub chunk - comment, verbatim
    Creator         : string;                 // ISFT sub chunk - tool used to make the font, verbatiom
  end;


  TSF2PresetHeader = packed record            // --- SF2 --- from <LIST><sdta>
  public
    PresetNameA  : array[ 1 .. 20] of AnsiChar;
    Preset       : Word;
    Bank         : Word;
    PresetBafNdx : Word;
    Lib          : Cardinal;
    Genre        : Cardinal;
    Morphology   : Cardinal;
  public
    function  PresetName: string;
  end;
  TSF2PresetHeaders = array of TSF2PresetHeader; // --- SF2 --- from <LIST><sdta>


  TSF2PresetBag = packed record
  public
    GenNdx : Word;
    ModNdx : Word;
  end;
  TSF2PresetBags = array of TSF2PresetBag;


  TSF2ModList = packed record
  public
    ModSrcOper    : Word;
    ModDestOper   : Word;
    ModAmount     : Int16;
    ModAmtSrcOper : Word;
    ModTransOper  : Word;
  end;
  TSF2ModLists = array of TSF2ModList;


  TSF2GenList = packed record
  public
    GenOper   : Word;
    GenAmount : Word;
  end;
  TSF2GenLists = array of TSF2GenList;


  TSF2Inst = packed record
  public
    InstNameA    : array[ 1 .. 20] of AnsiChar;
    BagNdx       : Word;
  public
    function  InstName: string;
  end;
  TSF2Insts = array of TSF2Inst;


  TSF2InstBag = packed record
  public
    InstGenNdx : Word;
    InstModNdx : Word;
  end;
  TSF2InstBags = array of TSF2InstBag;


  TSF2Sample = packed record
  public
    SampleNameA     : array[ 1 .. 20] of AnsiChar;  //  0
    StartPoint      : Cardinal;                     // 20
    EndPoint        : Cardinal;                     // 24
    StartLoop       : Cardinal;                     // 28
    EndLoop         : Cardinal;                     // 32
    SampleRate      : Cardinal;                     // 36
    OriginalPitch   : Byte;                         // 40
    PitchCorrection : Int8;                         // 41
    SampleLink      : Word;                         // 42
    SampleType      : Word;                         // 44
  public                                            // 46 - ok
    function  SampleName: string;
  end;
  TSF2Samples = array of TSF2Sample;


  TRiffReaderWriter = class                   // ---  common ---
  private
    FFourCC   : TFourCC;
    FSize     : Cardinal;
    FChunks   : TList;
    FErrorStr : string;
    FIsValid  : Boolean;
  private
    procedure   AddError( const aMsg: string);
    procedure   AddErrorFmt( const aFmt: string; const anArgs: array of const);
  protected
    function    GetChunkCount: Integer;
    function    GetChunk      ( anIndex: Integer): TBytesStream;
    function    GetChunkName  ( anIndex: Integer): TFourCC;
    function    GetChunkByName( const anIndex: TFourCC): TBytesStream;
  public
    constructor Create;
    destructor  Destroy;                                                                                       override;
    procedure   Clear;
    function    LoadFromFile( const aFilename: string; aMaxSize: Cardinal): Boolean;                            virtual;
    procedure   SaveToFile  ( const aFilename: string);                                                         virtual;
    function    IndexOf ( aChunkFourCC: TFourCC): Integer;
    function    HasChunk( aChunkFourCC: TFourCC): Boolean;
    function    AddChunk( aChunkFourCC: TFourCC): TBytesStream;
    procedure   DeleteChunk( anIndex: Integer);                                                                overload;
    procedure   DeleteChunk( aChunkFourCC: TFourCC);                                                           overload;
  public
    property    ErrorStr                            : string       read FErrorStr;
    property    FourCC                              : TFourCC      read FFourCC       write FFourCC;
    property    Size                                : Cardinal     read FSize;
    property    ChunkCount                          : Integer      read GetChunkCount;
    property    Chunk      [       anIndex: Integer]: TBytesStream read GetChunk;
    property    ChunkName  [       anIndex: Integer]: TFourCC      read GetChunkName;
    property    ChunkByName[ const anIndex: TFourCC]: TBytesStream read GetChunkByName;
  end;


  TWaveFileReader = class            // ---  common ---
  private
    FIsValid          : Boolean;
    FAutoNormalize    : Boolean;
    FHasAcid          : Boolean;
    FHasSampleInfo    : Boolean;
    FHasSF2Info       : Boolean;
    FFormatInfo       : TFormatInfo;
    FAcidInfo         : TAcidInfo;
    FSampleInfo       : TSampleInfo;
    FFrameCount       : Cardinal;
    FLoopInfo         : array of TLoopInfo;
    FCueInfo          : array of TCueInfo;
    FCueNames         : array of TCueName;
    FSF2Info          : TSF2Info;
    FSF2PresetHeaders : TSF2PresetHeaders;
    FSF2PresetBags    : TSF2PresetBags;
    FSF2PModLists     : TSF2ModLists;
    FSF2PGenLists     : TSF2GenLists;
    FSF2Insts         : TSF2Insts;
    FSF2InstBags      : TSF2InstBags;
    FSF2IModLists     : TSF2ModLists;
    FSF2IGenLists     : TSF2GenLists;
    FSF2Samples       : TSF2Samples;
    FErrorStr         : string;
    FMinValue         : TSignal;
    FMaxValue         : TSignal;
  private
    function    GetStreamFormat : Word;
    function    GetDuration     : TSignal;
    function    GetChannelCount : Word;
    function    GetSampleRate   : Cardinal;
    function    GetSampleSize   : Word;
    function    GetBlockAlign   : Word;
    function    GetLoopInfoCount: Integer;
    function    GetCueInfoCount : Integer;
    function    GetCueNameCount : Integer;
    function    GetLoopInfo( anIndex: Integer ): TLoopInfo;
    function    GetCueInfo ( anIndex: Integer ): TCueInfo;
    function    GetCueName ( anId   : Cardinal): TCueName;
    function    ReadStreamStr( const aStream: TMemoryStream; aCount: Cardinal): string;
    function    ParseData( const aReader: TRiffReaderWriter; var aWaveData: TWaveData): Boolean;
    procedure   AddCueName( const anInfo: TLoopInfo; const aName: string);                                           overload;
    procedure   AddCueName( const anInfo: TCueInfo ; aSampleLength: Cardinal; const aName: string; aType: TCueType); overload;
    procedure   AddCue   ( anId: Word; aPosition, aSampleLength: Cardinal; aType: TCueType; const aName: string);
    procedure   AddLabel ( anId: Word; aPosition: Cardinal; const aName: string);
    procedure   AddRegion( anId: Word; aPosition, aSampleLength: Cardinal; const aName: string);
    procedure   AddLoop  ( anId: Word; aPosition, aSampleLength: Cardinal; const aName: string);
    procedure   AddError( const aMsg: string);
    procedure   AddErrorFmt( const aFmt: string; const anArgs: array of const);
  public
    function    ReadData( const aReader: TRiffReaderWriter; var aWaveData: TWaveData): Boolean;
    procedure   NormalizeData( const aWaveData: TWaveData);
    function    HasLoops: Boolean;
    function    Dump: string;
  public
    property    IsValid                      : Boolean     read FIsValid;
    property    AutoNormalize                : Boolean     read FAutoNormalize   write FAutoNormalize;
    property    HasAcid                      : Boolean     read FHasAcid;
    property    HasSampleInfo                : Boolean     read FHasSampleInfo;
    property    StreamFormat                 : Word        read GetStreamFormat;
    property    Duration                     : TSignal     read GetDuration;
    property    ChannelCount                 : Word        read GetChannelCount;
    property    SampleRate                   : Cardinal    read GetSampleRate;
    property    SampleSize                   : Word        read GetSampleSize;
    property    BlockAlign                   : Word        read GetBlockAlign;
    property    FormatInfo                   : TFormatInfo read FFormatInfo;
    property    AcidInfo                     : TAcidInfo   read FAcidInfo;
    property    SampleInfo                   : TSampleInfo read FSampleInfo;
    property    LoopInfoCount                : Integer     read GetLoopInfoCount;
    property    CueInfoCount                 : Integer     read GetCueInfoCount;
    property    CueNameCount                 : Integer     read GetCueNameCount;
    property    FrameCount                   : Cardinal    read FFrameCount;
    property    LoopInfo[ anIndex: Integer ] : TLoopInfo   read GetLoopInfo;
    property    CueInfo [ anIndex: Integer ] : TCueInfo    read GetCueInfo;
    property    CueName [ anId   : Cardinal] : TCueName    read GetCueName;
    property    ErrorStr                     : string      read FErrorStr;
    property    MinValue                     : TSignal     read FMinValue;
    property    MaxValue                     : TSignal     read FMaxValue;
  end;


implementation


    function    MakeSampleRate( aValue: TExtArray): Cardinal;
    const
      DEFAULT_RATE = 44100;                // Default sample rate when Result could not be calculated
      EXP_BIAS     = 16383;                // For number conversion
      INT64_BITS   = 8 * SizeOf( UInt64);  // Number of bits in a 64 bit number

      function ExtSize: Integer;           // We use a function here as SizeOf( Extended) is a constant, resulting in
      begin                                // some code to not be compiled .. so it can not all be tested ... and there
        Result := SizeOf( Extended);       // will be compiler complaints about unused variables as well then. Sooo ..
      end;                                 // this is just to fool the compiler a bit.

    type
      TCExpo = record                      // Exponent conversion record
      case Byte of
        0 : ( E : Word);
        1 : ( B : array[ 0 .. 1] of Byte);
      end;
      TCMant = record                      // Mantissa conversion record
      case Byte of
        0 : ( M : UInt64);
        1 : ( B : array[ 0 .. 7] of Byte);
      end;
      TCExt = record                       // Plain Extended conversion record
      case Byte of
        0 : ( E : Extended);
        1 : ( B : array[ 0 .. 9] of Byte);
      end;
    var
      Expo   : Word;                       // Byte swapped exponent
      Mant   : UInt64;                     // Byte swapped mantissa with explicit 1 or 0 prepended (the IntBit)
      CExpo  : TCExpo;                     // Exponent conversion
      CMant  : TCMant;                     // Mantissa conversion
      Sign   : Integer;                    // The Extended number's Sign bit
      IntBit : Integer;                    // The value of the explicit zero or one mantissa prefix
      Ext    : TCExt;                      // Extended conversion
    begin
      Result := 0;

      if   ExtSize <> 10
      then begin                                     // Extended type is not 10 bytes in size, do a manual conversion

        // Testcase was .. yeah the only one .. : SampleRate = [ 40 0e bb 80 00 00 00 00 00 00 ] -> Result = 48000

        CMant.B[ 7] := aValue[ 2];                   // NO IntBit clear on mantissa, and store stuff byte swapped
        CMant.B[ 6] := aValue[ 3];
        CMant.B[ 5] := aValue[ 4];
        CMant.B[ 4] := aValue[ 5];
        CMant.B[ 3] := aValue[ 6];
        CMant.B[ 2] := aValue[ 7];
        CMant.B[ 1] := aValue[ 8];
        CMant.B[ 0] := aValue[ 9];

        CExpo.B[ 1] := aValue[ 0] and $7f;
        CExpo.B[ 0] := aValue[ 1];

        Mant        := CMant.M;                      // Byte swapped Mantissa
        Expo        := CExpo.E;                      // Byte swapped Exponent
        Sign        := aValue[ 0] shr 7;             // Get Sign
        IntBit      := aValue[ 2] shr 7;             // Get IntBit

        if  ( Expo = 0)                              // Zero as a special case, not interested in the sign here
        and ( Mant = 0)                              // Mant includes IntBit, so thats tested to be zero as well here
        then Result := 0                             // IntBit = 1 -> some denormal stuff
        else begin
          if   ( Expo   <> 0    )                    // Exp = 0 and Intbit = 1 -> pseaudo denormal, not handled
          and  ( Expo   <> $7fff)                    // Expo all ones: all cases of Nan, Inf, Signaling or just invalid
          and  ( IntBit <> 0    )                    // Intbit = 0 for regular numbers -> Denormal value, not supported here
          and  ( Sign   =  0    )                    // Sign = 0 -> Positive, not interested in negative sample rates here
          then Result := Mant shr ( INT64_BITS - Expo + EXP_BIAS - 1);
        end;
      end
      else begin                                     // Extended type is 10 bytes in size, some compiler magix will do
        Move( aValue[ 0], Ext.B[ 0], 10);
        Result := Round( ByteSwap( Ext.E));
      end;

      if   Result = 0                                // Sample rate 0 means .. either not specified or not computable
      then Result := DEFAULT_RATE;                   // return the default rate then
    end;




{ ========
  TCueName = record                           // ---  common ---
  public
    Id      : Cardinal;
    Name    : string;
    CueType : TCueType;
  public
}

    function    TCueName.CueTypeName: string;
    begin
      case CueType of
        ctLabel  : Result := 'ctLabel' ;
        ctRegion : Result := 'ctRegion';
        ctLoop   : Result := 'ctLoop'  ;
        else       Result := '<ctUnknown>';
      end
    end;


{ ========
  TCommInfo = packed record                   // ---  AIF ---
  public
    numChannels     : SmallInt;               //  0 | Audio channels
    numSampleFrames : Cardinal;               //  2 | Samples / channel
    SampleSize      : SmallInt;               //  6 | Bits / Sample
    SampleRate      : TExtArray;              //  8 | SampleFrams / sec - documented as extended .. but seems to not be IEEE ...
//  public                                    // 18
}

    procedure   TCommInfo.DoByteSwaps;
    begin
      numChannels     := ByteSwap( numChannels);
      numSampleFrames := ByteSwap( numSampleFrames);
      SampleSize      := ByteSwap( SampleSize);
    end;


    function    TCommInfo.GetSampleRate: Cardinal;
    begin
      Result := MakeSampleRate( SampleRate);
    end;


{ ========
  TCommInfoExt = packed record
  public
    numChannels     : SmallInt;               //  0 | Audio channels
    numSampleFrames : Cardinal;               //  2 | Samples / channel
    SampleSize      : SmallInt;               //  6 | Bits / Sample
    SampleRate      : TExtArray;              //  8 | SampleFrams / sec - documented as extended .. but seems to not be IEEE ...
    CompressionType : TFourCC;                // 18 | The compression type used
    StrLength       : Byte;                   // 22 The amount of string data that will follo
 // StringData      : array[ StrLength] of AnsiChar
  public                                      // 23 + StrLength zero padded to even length
}

    procedure   TCommInfoExt.DoByteSwaps;
    begin
      numChannels     := ByteSwap( numChannels);
      numSampleFrames := ByteSwap( numSampleFrames);
      SampleSize      := ByteSwap( SampleSize);
    end;


    function    TCommInfoExt.GetSampleRate: Cardinal;
    begin
      Result := MakeSampleRate( SampleRate);
    end;



const

  fmtPCM   = 1;
  fmtFloat = 3;


type

  PChunkInfo = ^TChunkInfo;
  TChunkInfo = record
    FourCC : TFourCC;
    Stream : TBytesStream;
  end;

  TRiffHeader = packed record
    Signature : TFourCC;
    Size      : Cardinal;
  end;


{ ========
  TSF2PresetHeader = packed record            // --- SF2 --- from <LIST><sdta>
  public
    PresetNameA  : array[ 1 .. 20] of AnsiChar;
    Preset       : Word;
    Bank         : Word;
    PresetBafNdx : Word;
    Lib          : Cardinal;
    Genre        : Cardinal;
    Morphology   : Cardinal;
  public
}


    function  TSF2PresetHeader.PresetName: string;
    var
      i : Integer;
      C : Char;
    begin
      Result := '';

      for i := 1 to 20
      do begin
        C := Char( PresetNameA[ i]);

        if C = #0
        then Break
        else Result := Result + C;
      end;
    end;


{ ========
  TSF2Inst = packed record
  public
    InstNameA    : array[ 1 .. 20] of AnsiChar;
    BagNdx       : Word;
  public
}

    function  TSF2Inst.InstName: string;
    var
      i : Integer;
      C : Char;
    begin
      Result := '';

      for i := 1 to 20
      do begin
        C := Char( InstNameA[ i]);

        if C = #0
        then Break
        else Result := Result + C;
      end;
    end;


{ ========
  TSF2Sample = packed record
  public
    SampleNameA     : array[ 1 .. 20] of AnsiChar;  //  0
    StartPoint      : Cardinal;                     // 20
    EndPoint        : Cardinal;                     // 24
    StartLoop       : Cardinal;                     // 28
    EndLoop         : Cardinal;                     // 32
    SampleRate      : Cardinal;                     // 36
    OriginalPitch   : Byte;                         // 40
    PitchCorrection : Int8;                         // 41
    SampleLink      : Word;                         // 42
    SampleType      : Word;                         // 44
  public                                            // 46 - ok
}

    function  TSF2Sample.SampleName: string;
    var
      i : Integer;
      C : Char;
    begin
      Result := '';

      for i := 1 to 20
      do begin
        C := Char( SampleNameA[ i]);

        if C = #0
        then Break
        else Result := Result + C;
      end;
    end;


{ ========
  TRiffReaderWriter = class
  private
    FFourCC : TFourCC;
    FSize   : Cardinal;
    FChunks : TList;
  public
    property    FourCC                              : TFourCC      read FFourCC       write FFourCC;
    property    Size                                : Cardinal     read FSize;
    property    ChunkCount                          : Integer      read GetChunkCount;
    property    Chunk      [       anIndex: Integer]: TBytesStream read GetChunk;
    property    ChunkName  [       anIndex: Integer]: TFourCC      read GetChunkName;
    property    ChunkByName[ const anIndex: TFourCC]: TBytesStream read GetChunkByName;
  private
}

    procedure   TRiffReaderWriter.AddError( const aMsg: string);
    begin
      if   FErrorStr = ''
      then FErrorStr :=                    aMsg
      else FErrorStr := FErrorStr + '; ' + aMsg;
    end;


    procedure   TRiffReaderWriter.AddErrorFmt( const aFmt: string; const anArgs: array of const);
    begin
      AddError( Format( aFmt, anArgs, AppLocale));;
    end;


//  protected

    function    TRiffReaderWriter.GetChunkCount: Integer;
    begin
      Result := FChunks.Count;
    end;


    function    TRiffReaderWriter.GetChunk( anIndex: Integer): TBytesStream;
    begin
      if   ( anIndex >= 0)
      and  ( anIndex < ChunkCount)
      then Result := PChunkInfo(FChunks[ anIndex])^.Stream
      else Result := nil;
    end;


    function    TRiffReaderWriter.GetChunkName( anIndex: integer): TFourCC;
    begin
      if   ( anIndex >= 0)
      and  ( anIndex < ChunkCount)
      then Result := PChunkInfo( FChunks[ anIndex])^.FourCC
      else Result := '*err';
    end;


    function    TRiffReaderWriter.GetChunkByName( const anIndex: TFourCC): TBytesStream;
    var
      p : Integer;
    begin
      p := IndexOf( anIndex);

      if p >= 0
      then Result := Chunk[ p]
      else Result := nil;
    end;


//  public

    constructor TRiffReaderWriter.Create;
    begin
      inherited;
      FChunks := TList.Create;
      Clear;
    end;


    destructor  TRiffReaderWriter.Destroy; // override;
    begin
      Clear;
      FreeAndNil( FChunks);
      inherited;
    end;


    procedure   TRiffReaderWriter.Clear;
    begin
      while ChunkCount > 0
      do    DeleteChunk( 0);

      FFourCC  := #0#0#0#0;
      FIsValid := False;
    end;


    function    TRiffReaderWriter.LoadFromFile( const aFilename: string; aMaxSize: Cardinal): Boolean; // virtual;
    var
      aHeader : TRiffHeader;
      aFourCC : TFourCC;
      aSize   : cardinal;
      aStream : TBytesStream;
    begin
      Clear;
      Result  := False;
      aStream := TBytesStream.Create;

      try
        aStream.LoadFromFile( aFilename);

        if   aStream.Size > aMaxSize
        then begin
          Clear;
          AddError( Format( 'File > %d MiB', [ aMaxSize div ( 1024 * 1024)], AppLocale));
          Exit;
        end;

        if   aStream.Read( aHeader, SizeOf( aHeader)) <> SizeOf( aHeader)
        then AddError( 'no RIFF header');

        if   SameText( string( aHeader.Signature), 'RIFF')   // Window RIFF file, for WAVs
        then begin
          FSize := aHeader.Size;

          if   aStream.Read( FFourCC, SizeOf( FFourCC)) <> SizeOf( FFourCC)
          then AddError( 'no RIFF type');

          while aStream.Position < aStream.Size - SizeOf( aFourCC) - SizeOf( aSize) // Be forgiving for bad RIFFs
          do begin
            if   aStream.Read( aFourCC, SizeOf( aFourCC)) <> SizeOf( aFourCC)
            then AddError( 'no CHUNK type');

            if   aStream.Read( aSize  , SizeOf( aSize)) <> SizeOf( aSize)
            then AddError( 'no CHUNK size');

            with AddChunk( aFourCC)
            do begin
              SetSize( aSize);
          {$ifdef win32}
              Write( Pointer( Cardinal( aStream.Memory) + aStream.Position)^, aSize);
          {$else}
              Write( Pointer( Int64( aStream.Memory) + aStream.Position)^, aSize);
          {$endif}
              aStream.Seek( aSize, soCurrent);
            end;
          end;
        end
        else if SameText( string( aHeader.Signature), 'FORM')  // Apple FORM file, for AIFs
        then begin
          FSize := ByteSwap( aHeader.Size);

          if   aStream.Read( FFourCC, SizeOf( FFourCC)) <> SizeOf( FFourCC)
          then AddError( 'no FORM type');

          while aStream.Position < aStream.Size - SizeOf( aFourCC) - SizeOf( aSize) // Be forgiving for bad RIFFs
          do begin
            if   aStream.Read( aFourCC, SizeOf( aFourCC)) <> SizeOf( aFourCC)
            then AddError( 'no CHUNK type');

            if   aStream.Read( aSize  , SizeOf( aSize)) <> SizeOf( aSize)
            then AddError( 'no CHUNK size');

            with AddChunk( aFourCC)
            do begin
              aSize := ByteSwap( aSize);
              SetSize( aSize);
          {$ifdef win32}
              Write( Pointer( Cardinal( aStream.Memory) + aStream.Position)^, aSize);
          {$else}
              Write( Pointer( Int64( aStream.Memory) + aStream.Position)^, aSize);
          {$endif}
              aStream.Seek( aSize, soCurrent);
            end;
          end;
        end;

        Result := True;
      finally
        aStream.DisposeOf;
      end;
    end;


    procedure   TRiffReaderWriter.SaveToFile( const aFilename: string); // virtual;
    var
      aHeader     : TRiffHeader;
      aFileStream : TFileStream;
      aSize       : Cardinal;
      i           : Integer;
    begin
      aHeader.Signature := 'RIFF';
      aHeader.Size      := 0;

      for i := 0 to FChunks.Count - 1
      do aHeader.Size := aHeader.Size + PChunkInfo( FChunks[ i])^.Stream.Size;

      aHeader.Size := aHeader.Size + Cardinal(( FChunks.Count + 1)) * Cardinal( SizeOf( TFourCC));
      aFileStream  := TFileStream.Create( aFilename, fmCreate or fmOpenWrite or fmShareExclusive);
      FSize        := aHeader.Size;

      try
        if   aFileStream.Write( aHeader, SizeOf( aHeader)) <> SizeOf( aHeader)
        then AddError( 'Can''t write RIFF header');

        if    aFileStream.Write( FFourCC, SizeOf( FFourCC)) <> SizeOf( FFourCC)
        then AddError( 'Can''t write RIFF type');

        for i := 0 to FChunks.Count - 1
        do begin
          if   aFileStream.Write( PChunkInfo( FChunks[ i])^.FourCC, SizeOf( PChunkInfo( FChunks[ i])^.FourCC)) <> SizeOf( PChunkInfo( FChunks[ i])^.FourCC)
          then AddError(  'Can''t write chunk type');

          aSize := PChunkInfo( FChunks[ i])^.Stream.Size;

          if   aFileStream.Write( aSize, SizeOf( aSize)) <> SizeOf( aSize)
          then AddError(  'Can''t write chunk size');

          if   Cardinal( aFileStream.Write( PChunkInfo( FChunks[ i])^.Stream.Memory^, aSize)) <> aSize
          then AddError( 'Can''t write chunk data');
        end;
      finally
        aFileStream.DisposeOf;
      end;
    end;


    function    TRiffReaderWriter.IndexOf( aChunkFourCC: TFourCC): Integer;
    var
      i : Integer;
    begin
      Result := -1;

      for i := 0 to FChunks.Count - 1
      do begin
        if   PChunkInfo( FChunks[ i])^.FourCC = aChunkFourCC
        then begin
          Result := i;
          Break;
        end;
      end;
    end;


    function    TRiffReaderWriter.HasChunk( aChunkFourCC: TFourCC): Boolean;
    begin
      Result := IndexOf( aChunkFourCC) >= 0;
    end;


    function    TRiffReaderWriter.AddChunk( aChunkFourCC: TFourCC): TBytesStream;
    var
      pChunk: PChunkInfo;
    begin
      if   ( not SameText( string( aChunkFourCC), 'LIST'))
      and  ( not SameText( string( aChunkFourCC), 'JUNK'))
      and  HasChunk( aChunkFourCC)
      then begin
        Result := nil;
        AddErrorFmt( 'Chunk ''%s'' already exists', [ string( aChunkFourCC)]);
        Exit;
      end;

      pChunk := AllocMem( SizeOf( TChunkInfo));
      FChunks.Add( pChunk);
      pChunk^.FourCC := aChunkFourCC;
      pChunk^.Stream := TBytesStream.Create;
      Result         := pChunk^.Stream;
    end;


    procedure   TRiffReaderWriter.DeleteChunk( anIndex: Integer); // overload;
    var
      pChunk: PChunkInfo;
    begin
      if   ( anIndex >= 0)
      and  ( anIndex < ChunkCount)
      then begin
        pChunk := FChunks[ anIndex];
        pChunk^.Stream.DisposeOf;
        FreeMem( pChunk);
        FChunks.Delete( anIndex);
      end;
    end;


    procedure   TRiffReaderWriter.DeleteChunk( aChunkFourCC: TFourCC); // overload;
    var
      p : integer;
    begin
      p := IndexOf( aChunkFourCC);

      if   p >= 0
      then DeleteChunk( p);
    end;


{ ========
  TWaveFileReader = class
  private
    FIsValid          : Boolean;
    FAutoNormalize    : Boolean;
    FHasAcid          : Boolean;
    FHasSampleInfo    : Boolean;
    FHasSF2Info       : Boolean;
    FFormatInfo       : TFormatInfo;
    FAcidInfo         : TAcidInfo;
    FSampleInfo       : TSampleInfo;
    FFrameCount       : Cardinal;
    FLoopInfo         : array of TLoopInfo;
    FCueInfo          : array of TCueInfo;
    FCueNames         : array of TCueName;
    FSF2Info          : TSF2Info;
    FSF2PresetHeaders : TSF2PresetHeaders;
    FSF2PresetBags    : TSF2PresetBags;
    FSF2PModLists     : TSF2ModLists;
    FSF2PGenLists     : TSF2GenLists;
    FSF2Insts         : TSF2Insts;
    FSF2InstBags      : TSF2InstBags;
    FSF2IModLists     : TSF2ModLists;
    FSF2IGenLists     : TSF2GenLists;
    FSF2Samples       : TSF2Samples;
    FErrorStr         : string;
    FMinValue         : TSignal;
    FMaxValue         : TSignal;
  public
    property    IsValid                      : Boolean     read FIsValid;
    property    AutoNormalize                : Boolean     read FAutoNormalize   write FAutoNormalize;
    property    HasAcid                      : Boolean     read FHasAcid;
    property    HasSampleInfo                : Boolean     read FHasSampleInfo;
    property    StreamFormat                 : Word        read GetStreamFormat;
    property    Duration                     : TSignal     read GetDuration;
    property    ChannelCount                 : Word        read GetChannelCount;
    property    SampleRate                   : Cardinal    read GetSampleRate;
    property    SampleSize                   : Word        read GetSampleSize;
    property    BlockAlign                   : Word        read GetBlockAlign;
    property    FormatInfo                   : TFormatInfo read FFormatInfo;
    property    AcidInfo                     : TAcidInfo   read FAcidInfo;
    property    SampleInfo                   : TSampleInfo read FSampleInfo;
    property    LoopInfoCount                : Integer     read GetLoopInfoCount;
    property    CueInfoCount                 : Integer     read GetCueInfoCount;
    property    CueNameCount                 : Integer     read GetCueNameCount;
    property    FrameCount                   : Cardinal    read FFrameCount;
    property    LoopInfo[ anIndex: Integer ] : TLoopInfo   read GetLoopInfo;
    property    CueInfo [ anIndex: Integer ] : TCueInfo    read GetCueInfo;
    property    CueName [ anId   : Cardinal] : TCueName    read GetCueName;
    property    ErrorStr                     : string      read FErrorStr;
    property    MinValue                     : TSignal     read FMinValue;
    property    MaxValue                     : TSignal     read FMaxValue;
  private
}

    function    TWaveFileReader.GetStreamFormat: Word;
    begin
      Result := FormatInfo.WaveFormat;
    end;


    function    TWaveFileReader.GetDuration: TSignal; // Duration in seconds
    begin
      if   SampleRate > 0
      then Result := FrameCount / SampleRate
      else Result := 0;
    end;


    function    TWaveFileReader.GetChannelCount: Word;
    begin
      Result := FormatInfo.ChannelCount;
    end;


    function    TWaveFileReader.GetSampleRate: Cardinal;
    begin
      Result := FormatInfo.SampleRate;
    end;


    function    TWaveFileReader.GetSampleSize: Word;
    begin
      Result := BlockAlign div ChannelCount;
    end;


    function    TWaveFileReader.GetBlockAlign: Word;
    begin
      Result := FormatInfo.BlockAlign;
    end;


    function    TWaveFileReader.GetLoopInfoCount: Integer;
    begin
      Result := Length( FLoopInfo);
    end;


    function    TWaveFileReader.GetCueInfoCount : Integer;
    begin
      Result := Length( FCueInfo);
    end;


    function    TWaveFileReader.GetCueNameCount: Integer;
    begin
      Result := Length( FCueNames);
    end;


    function    TWaveFileReader.GetLoopInfo( anIndex: Integer): TLoopInfo;
    begin
      FillChar( Result, SizeOf( Result), 0);

      if   IsValid
      and  ( anIndex >= 0)
      and  ( anIndex < LoopInfoCount)
      then Result := FLoopInfo[ anIndex];
    end;


    function    TWaveFileReader.GetCueInfo ( anIndex: Integer): TCueInfo;
    begin
      FillChar( Result, SizeOf( Result), 0);

      if   IsValid
      and  ( anIndex >= 0)
      and  ( anIndex < CueInfoCount)
      then Result := FCueInfo[ anIndex];
    end;


    function    TWaveFileReader.GetCueName( anId: Cardinal): TCueName;
    var
      i : Integer;
    begin
      Result.Id      := 0 ;         // Can't use FillChar here ... string reference count not kept properly then
      Result.Name    := '';         // and also ... it's a bit odd to init strings with FillChar anyway.
      Result.CueType := ctUnknown;

      if   IsValid
      then begin
        for i := 0 to CueNameCount - 1
        do begin
          if   FCueNames[ i].Id = anId
          then begin
            Result := FCueNames[ i];
            Break;
          end;
        end;
      end;
    end;


    function    TWaveFileReader.ReadStreamStr( const aStream: TMemoryStream; aCount: Cardinal): string;
    var
      B : AnsiChar;
      i : Integer;
    begin
      Result := '';

      for i := 0 to aCount - 1
      do begin
        aStream.Read( B, SizeOf( B));

        if   B <> #0
        then Result := Result + Char( B);
      end;

      if   Odd( aCount)                             // Handle string padding
      then aStream.Seek( 1, soCurrent);
    end;


    function    TWaveFileReader.ParseData( const aReader: TRiffReaderWriter; var aWaveData: TWaveData): Boolean; // overload
    type
      TSample = record
        case Integer of
          0 : ( UInt8Val  : UInt8 );  // 8 bit audio with 128 as the center value
          1 : ( Int16Val  : Int16 );
          2 : ( Int24Val  : Int32 );
          3 : ( Int32Val  : Int32 );
          4 : ( Int64Val  : Int64 );
          5 : ( SingleVal : Single);
          6 : ( DoubleVal : Double);
          7 : ( ArrayVal  : array[ 0 .. 7] of Byte);
        end;
    var
      aStream          : TBytesStream;
      aCount           : Cardinal;
      aFourCC          : TFourCC;
      aSample          : TSample;
      anId             : Cardinal;
      S                : string;
      T                : string;
      i                : Integer;
      j                : Integer;
      dummyw           : Word;
      aCommInfoExt     : TCommInfoExt;
      aMarkerCount     : Word;
      aMarkerId        : Word;
      aMarkerPosition  : Cardinal;
      anStrLen         : Byte;
      anOffset         : Cardinal;
      aBlockSize       : Cardinal;
      aCompressFlag    : Boolean;
      aSowt            : Boolean;
      aSize            : Cardinal;
      IsFloat          : Boolean;
      SampleLength     : Cardinal;
      IntSample        : Int16;
      ByteSample       : UInt8;
      SampleValue      : Int32;
      FirstSampleRate  : Cardinal;
    begin
      Result    := False;
      FMinValue :=  1.0;
      FMaxValue := -1.0;
      FillChar ( FFormatInfo, SizeOf( FFormatInfo), 0);
      FillChar ( FAcidInfo  , SizeOf( FAcidInfo  ), 0);
      FillChar ( FSampleInfo, SizeOf( FSampleInfo), 0);
      SetLength( FLoopInfo  , 0);
      SetLength( FCueInfo   , 0);
      SetLength( FCueNames  , 0);
      SetLength( aWaveData  , 0);
      FHasAcid        := False;
      FHasSampleInfo  := False;
      FirstSampleRate := 0;

      // ---- WAVE ----

      if   SameText( string( aReader.FourCC), 'WAVE')             // Accept WAVE files
      and  aReader.HasChunk( 'data')                              // that have data
      and  aReader.HasChunk( 'fmt ')                              // and have a format description
      then begin
        aStream := aReader.ChunkByName[ 'fmt '];
        aStream.Seek( 0, soBeginning);
        aStream.Read( FFormatInfo, SizeOf( FFormatInfo));         // Not reading possible ExtraCount bytes, but ExtraCount item is read

        if   (( StreamFormat = fmtPCM  ) and ( SampleSize in [ 1, 2, 3, 4, 8])) // Accept 8, 16, 24, 32 or 64 bit PCM   format
        or   (( StreamFormat = fmtFloat) and ( SampleSize in [          4, 8])) // Accept            32 or 64 bit Float format
        then begin
          if   aReader.HasChunk( 'acid')                          // Parse out possible 'acid' chunk
          then begin
            aStream := aReader.ChunkByName[ 'acid'];
            aStream.Seek( 0, soBeginning);
            aStream.Read( FAcidInfo, SizeOf( FAcidInfo));
            FHasAcid := True;
          end;

          if   aReader.HasChunk( 'smpl')                          // Parse out possible 'smpl' chunk
          then begin
            aStream := aReader.ChunkByName[ 'smpl'];
            aStream.Seek( 0, soBeginning);
            aStream.Read( FSampleInfo, SizeOf( FSampleInfo));
            SetLength( FLoopInfo, FSampleInfo.SampleLoops);

            if   FSampleInfo.SampleLoops > 0
            then aStream.Read( FLoopInfo[ 0], FSampleInfo.SampleLoops * SizeOf( TLoopInfo));

            for i := 0 to FSampleInfo.SampleLoops - 1
            do AddCueName( FLoopInfo[ i], Format( 'loop %d', [ i + 1], AppLocale)); // Note: there are no native loop names in the 'cue ' chunk

            FHasSampleInfo := True;
          end;

          if   aReader.HasChunk( 'cue ')                          // Parse out possible 'cue ' chunk
          then begin
            aStream := aReader.ChunkByName[ 'cue '];
            aStream.Seek( 0, soBeginning);
            aStream.Read( aCount, SizeOf( aCount));
            SetLength( FCueInfo , aCount);

            if   aCount > 0
            then begin
              aStream.Read( FCueInfo[ 0], aCount * SizeOf( TCueInfo));

              for i := 0 to CueInfoCount - 1                      // Generate default names for in case no names are found later on
              do AddCueName( FCueInfo[ i], 0, Format( 'cue %d', [ i + 1], AppLocale), ctUnknown);
            end;
          end;

          if   aReader.HasChunk( 'LIST')                          // Try to find names and types for labels & regions, there can be more than one LIST chunk
          then begin
            for i := 0 to aReader.ChunkCount - 1
            do begin
              if   SameText( string( aReader.ChunkName[ i]), 'LIST')
              then begin
                aStream := aReader.Chunk[ i];
                aStream.Seek( 0, soBeginning);
                aStream.Read( aFourCC, SizeOf( aFourCC));

                if   SameText( string( aFourCC), 'adtl')          // For labels and ltxt items
                then begin
                  while aStream.Position < aStream.Size
                  do begin
                    aStream.Read( aFourCC, SizeOf( aFourCC));
                    aStream.Read( aCount , SizeOf( aCount) );

                    if   SameText( string( aFourCC), 'labl')      // we handle 'labl' sub-chunks here
                    then begin
                      aStream.Read( anId, SizeOf( anId));
                      S := ReadStreamStr( aStream, aCount);       // todo: check if padding handled correctly here

                      for j := 0 to Length( FCueNames) - 1
                      do begin
                        if   FCueNames[ j].Id = anId
                        then begin
                          FCueNames[ j].Name         := S;
                          FCueNames[ j].CueType      := ctLabel;
                          FCueNames[ j].SampleLength := 0;
                          Break;
                        end;
                      end;
                    end
                    else if SameText( string( aFourCC), 'ltxt')   // we handle 'ltxt' sub-chunks here
                    then begin
                      aStream.Read( anId, SizeOf( anId));
                      aStream.Read( SampleLength, SizeOf( SampleLength)); // dwSampleLength
                      aStream.Read( aFourCC     , SizeOf( aFourCC     )); // dwPurpose ... prolly 'regn' for a region ,,,
                      aStream.Read( dummyw      , SizeOf( dummyw      )); // wCountry
                      aStream.Read( dummyw      , SizeOf( dummyw      )); // wLanguage
                      aStream.Read( dummyw      , SizeOf( dummyw      )); // wDialect
                      aStream.Read( dummyw      , SizeOf( dummyw      )); // wCodePage
                      S := ReadStreamStr( aStream, aCount);               // todo: check if padding handled correctly here

                      if   SameText( string( aFourCC), 'rgn ')
                      then begin
                        for j := 0 to Length( FCueNames) - 1
                        do begin
                          if   FCueNames[ j].Id = anId
                          then begin
                            FCueNames[ j].Name         := S;
                            FCueNames[ j].CueType      := ctRegion;
                            FCueNames[ j].SampleLength := SampleLength;
                            Break;
                          end;
                        end;
                      end;
                    end
                    else aStream.Seek( aCount, soCurrent)       // Skip unhandled sub-chunk
                  end
                end
                else begin                                      // Unhandled LIST items .. hrmm .. prolly chunks too
                end
              end;
            end;
          end;

          aStream     := aReader.ChunkByName[ 'data'];            // Parse out the 'data' chunk - which has to be
          aCount      := aStream.Size;                            // pre-checked to be present, all data will be
          FFrameCount := aCount div BlockAlign;                   // converted to TSignal values, such that a user needs
          aStream.Seek( 0, soBeginning);                          // not know about the underlying wave encoding.
          SetLength( aWaveData, ChannelCount);

          for i := 0 to ChannelCount - 1
          do SetLength( aWaveData[ i], FFrameCount);

          for i := 0 to FFrameCount - 1
          do begin
            for j := 0 to ChannelCount - 1
            do begin
              aStream.Read( aSample, SampleSize);

              case StreamFormat of

                fmtPCM :

                  begin
                    case SampleSize of
                      1 : aWaveData[ j, i] := 1.0 * ( Integer( aSample.UInt8Val) - 128) / Int8.MaxValue;
                      2 : aWaveData[ j, i] := 1.0 * aSample.Int16Val / Int16.MaxValue;
                      3 : aWaveData[ j, i] := 1.0 * ( aSample.Int24Val shl 8) / Int32.MaxValue;
                      4 : aWaveData[ j, i] := 1.0 * aSample.Int32Val / Int32.MaxValue;
                      8 : aWaveData[ j, i] := 1.0 * aSample.Int64Val / Int64.MaxValue;
                    end;
                  end;

                fmtFloat :

                  begin
                    case SampleSize of
                      4 : aWaveData[ j, i] := aSample.SingleVal;
                      8 : aWaveData[ j, i] := aSample.DoubleVal;
                    end;
                  end;

              end;

              if   aWaveData[ j, i] < FMinValue
              then FMinValue := aWaveData[ j, i];

              if   aWaveData[ j, i] > FMaxValue
              then FMaxValue := aWaveData[ j, i];
            end;
          end;

          Result := True;
        end
        else AddError( 'Unsupported combination of SampleSize and WaveFormat - int 8, 16, 32, 64 and float 32, 64 supported');
      end

      // ---- AIFF, AIFC ----

      else if
           (
                SameText( string( aReader.FourCC), 'AIFF')        // Accept AIFF files (uncompressed files, not AIFC)
             or SameText( string( aReader.FourCC), 'AIFC')
           )
           and  aReader.HasChunk( 'SSND')                         // that have data
           and  aReader.HasChunk( 'COMM')                         // and have a format description
      then begin
        aCompressFlag := False;
        aSowt         := False;
        IsFloat       := False;
        aStream       := aReader.ChunkByName[ 'COMM'];

        if   SameText( string( aReader.FourCC), 'AIFF')
        then begin
          aStream.Seek( 0, soBeginning);
          aStream.Read( aCommInfoExt, SizeOf( TCommInfo));        // Just read what can be available .. last fields will be empty
          aCommInfoExt.DoByteSwaps;
        end
        else if SameText( string( aReader.FourCC), 'AIFC')
        then begin
          aStream.Seek( 0, soBeginning);

          if aStream.Size < SizeOf( aCommInfoExt)
          then begin
            aStream.Read( aCommInfoExt, SizeOf( TCommInfo));      // Try to read as if plain AIFF
            aReader.FourCC := 'AIFF';                             // Signal it to actually be that .. AIFF
          end
          else begin                                              // Read as actual AIFC
            aCompressFlag := True;                                // Compressed by default ...

            aStream.Read( aCommInfoExt, SizeOf( aCommInfoExt));

            if      SameText( string( aCommInfoExt.CompressionType), 'NONE')
            then aCompressFlag := False
            else if SameText( string( aCommInfoExt.CompressionType), 'sowt')
            then begin
              aCompressFlag := False;
              aSowt         := True;
            end
            else if SameText( string( aCommInfoExt.CompressionType), 'fl32') // Handles FL32 too
            then begin
              aCompressFlag := False;
              aSowt         := True;
              IsFloat       := True;
            end
            else if SameText( string( aCommInfoExt.CompressionType), 'fl64')
            then begin
              aCompressFlag := False;
              aSowt         := True;
              IsFloat       := True;
            end
          end;

          aCommInfoExt.DoByteSwaps;
        end;

        if   IsFloat
        then FFormatInfo.WaveFormat   := fmtFloat
        else FFormatInfo.WaveFormat   := fmtPCM;

        FFormatInfo.ChannelCount := aCommInfoExt.numChannels;
        FFormatInfo.SampleRate   := aCommInfoExt.GetSampleRate;
        FFormatInfo.BitCount     := aCommInfoExt.SampleSize;
        FFormatInfo.BlockAlign   := ( FFormatInfo.ChannelCount * FFormatInfo.BitCount) div 8;
        FFormatInfo.StreamSpeed  := FFormatInfo.SampleRate * FFormatInfo.BlockAlign;
        FFormatInfo.ExtraCount   := 0;

        if   ( not aCompressFlag)                                                 // Accept uncompressed only
        and  (
                  (( StreamFormat = fmtPCM  ) and ( SampleSize in [ 1, 2, 4, 8])) // Accept 8, 16, 32 or 64 bit PCM   format
               or (( StreamFormat = fmtFloat) and ( SampleSize in [       4, 8])) // Accept        32 or 64 bit Float format
             )
        then begin
          if   aReader.HasChunk( 'MARK')
          then begin
            aStream := aReader.ChunkByName[ 'MARK'];
            aStream.Seek( 0, soBeginning);
            aStream.Read( aMarkerCount, SizeOf( aMarkerCount));
            aMarkerCount := ByteSwap( aMarkerCount);

            for i := 0 to aMarkerCount - 1
            do begin
              aStream.Read( aMarkerId      , SizeOf( aMarkerId      ));
              aStream.Read( aMarkerPosition, SizeOf( aMarkerPosition));
              aStream.Read( anStrLen       , SizeOf( anStrLen       ));
              aMarkerId       := ByteSwap( aMarkerId      );
              aMarkerPosition := ByteSwap( aMarkerPosition);
              S := ReadStreamStr( aStream, anStrLen);
              T := UpperCase( S);

              // Hmm .. how to see what type of marker it is ... label, region or loop ...
              //        and also .. what is the loop and region length ??
              //        assuming loop and region lengths to be zero for AIFx files ...

              if      Pos( 'LABEL' , T) > 0 then AddLabel ( aMarkerId, aMarkerPosition,    S)
              else if Pos( 'REGION', T) > 0 then AddRegion( aMarkerId, aMarkerPosition, 0, S)
              else if Pos( 'LOOP'  , T) > 0 then AddLoop  ( aMarkerId, aMarkerPosition, 0, S)
            end;
          end;

          aStream     := aReader.ChunkByName[ 'SSND'];            // Parse out the 'SSND' chunk - which has to be
          FFrameCount := aCommInfoExt.numSampleFrames;            // pre-checked to be present, all data will be
          aStream.Seek( 0, soBeginning);                          // converted to TSignal values, such that a user needs
          SetLength( aWaveData, ChannelCount);                    // not know about the underlying wave encoding.

          for i := 0 to ChannelCount - 1
          do SetLength( aWaveData[ i], FFrameCount);

          aStream.Read( anOffset  , SizeOf( anOffset  ));         // offset
          aStream.Read( aBlockSize, SizeOf( aBlockSize));         // blockSize
          aStream.Seek( anOffset, soCurrent);                     // Skip offset bytes from SoundData[]

          for i := 0 to FFrameCount - 1
          do begin
            for j := 0 to ChannelCount - 1
            do begin
              aStream.Read( aSample, SampleSize);

              case StreamFormat of

                fmtPCM :

                  if   aSowt
                  then begin                                      // Intel byte order   - little-endian
                    case SampleSize of
                      1 : aWaveData[ j, i] := 1.0 * ( Integer( aSample.UInt8Val) - 128) / Int8.MaxValue;
                      2 : aWaveData[ j, i] := 1.0 * aSample.Int16Val / Int16.MaxValue;
                      3 : aWaveData[ j, i] := 1.0 * ( aSample.Int24Val shl 8) / Int32.MaxValue;
                      4 : aWaveData[ j, i] := 1.0 * aSample.Int32Val / Int32.MaxValue;
                      8 : aWaveData[ j, i] := 1.0 * aSample.Int64Val / Int64.MaxValue;
                    end;
                  end
                  else begin                                      // Old Mac byte order - big-endian
                    case SampleSize of
                      1 : aWaveData[ j, i] := 1.0 * ( Integer( ByteSwap( aSample.UInt8Val) - 128)) / Int8.MaxValue;
                      2 : aWaveData[ j, i] := 1.0 * ByteSwap( aSample.Int16Val) / Int16.MaxValue;
                      3 : aWaveData[ j, i] := 1.0 * ( ByteSwap( aSample.Int24Val) shl 8) / Int32.MaxValue;
                      4 : aWaveData[ j, i] := 1.0 * ByteSwap( aSample.Int32Val) / Int32.MaxValue;
                      8 : aWaveData[ j, i] := 1.0 * ByteSwap( aSample.Int64Val) / Int64.MaxValue;
                    end;
                  end;

                fmtFloat :

                  if   aSowt
                  then begin                                      // Intel byte order   - little-endian
                    case SampleSize of
                      4 : aWaveData[ j, i] := aSample.SingleVal;
                      8 : aWaveData[ j, i] := aSample.DoubleVal;
                    end;
                  end
                  else begin                                      // Old Mac byte order - big-endian (should never happen I guess)
                    case SampleSize of
                      4 : aWaveData[ j, i] := ByteSwap( aSample.SingleVal);
                      8 : aWaveData[ j, i] := ByteSwap( aSample.DoubleVal);
                    end;
                  end;

              end;

              if   aWaveData[ j, i] < FMinValue
              then FMinValue := aWaveData[ j, i];

              if   aWaveData[ j, i] > FMaxValue
              then FMaxValue := aWaveData[ j, i];
            end;
          end;

          Result := True;
        end
        else AddError( 'Unsupported sample depth in aif file, supported are [ 8, 16, 32, 64]');
      end

      // ---- SF2 (Sound font) ----

      else if
           (
                SameText( string( aReader.FourCC), 'sfbk')        // Accept SF2 files
           )
      then begin
        SetLength( aWaveData, 0);

        if   aReader.HasChunk( 'LIST')                            // Try to find all the list data
        then begin
          for i := 0 to aReader.ChunkCount - 1
          do begin
            if   SameText( string( aReader.ChunkName[ i]), 'LIST')
            then begin
              aStream := aReader.Chunk[ i];
              aStream.Seek( 0, soBeginning);
              aStream.Read( aFourCC, SizeOf( aFourCC));

              if   SameText( string( aFourCC), 'INFO')            // INFO list
              then begin
                FHasSF2Info := True;                              // So we do have an INFO list

                with FSF2Info
                do begin
                  Valid           := True;    // True when data was acceptable
                  VersionMajor    := 0;       // ifil sub chunk
                  VersionMinor    := 0;       // ifil sub chunk
                  SoundEngine     := '';      // isng sub chunk
                  BankName        := '';      // INAM sub chunk
                  CreationDate    := '';      // ICRD sub chunk - verbatim, whatever the format that is in the file
                  Engineers       := '';      // IENG sub chunk - engineers - verbatim
                  ProductId       := '';      // IPRD sub chunk - the product the font was made for, verbatim
                  Copyright       := '';      // ICOP sub chunk - copyright, verbatim
                  Comment         := '';      // ICMT sub chunk - comment, verbatim
                  Creator         := '';      // ISFT sub chunk - tool used to make the font, verbatiom
                end;

                while aStream.Position < aStream.Size             // Read sub-chunks
                do begin
                  aStream.Read( aFourCC, SizeOf( aFourCC));
                  aStream.Read( aCount , SizeOf( aCount) );

                  if      SameText( string( aFourCC), 'ifil')     // ifil sub chunk
                  then begin
                    if   aCount = 4
                    then begin
                      aStream.Read( FSF2Info.VersionMajor, 2);
                      aStream.Read( FSF2Info.VersionMinor, 2);
                    end
                    else aStream.Seek( aCount, soCurrent);
                  end
                  else if SameText( string( aFourCC), 'isng')     // isng sub chunk
                  then FSF2Info.SoundEngine := ReadStreamStr( aStream, aCount)
                  else if SameText( string( aFourCC), 'INAM')     // INAM sub chunk
                  then FSF2Info.BankName := ReadStreamStr( aStream, aCount)
                  else if SameText( string( aFourCC), 'irom')     // irom sub chunk
                  then aStream.Seek( aCount, soCurrent)           // skip it
                  else if SameText( string( aFourCC), 'iver')     // iver sub chunk
                  then aStream.Seek( aCount, soCurrent)           // skip it
                  else if SameText( string( aFourCC), 'ICRD')     // ICRD sub chunk
                  then FSF2Info.CreationDate := ReadStreamStr( aStream, aCount)
                  else if SameText( string( aFourCC), 'IENG')     // IENG sub chunk
                  then FSF2Info.Engineers := ReadStreamStr( aStream, aCount)
                  else if SameText( string( aFourCC), 'IPRD')     // IPRD sub chunk
                  then FSF2Info.ProductId := ReadStreamStr( aStream, aCount)
                  else if SameText( string( aFourCC), 'ICOP')     // ICOP sub chunk
                  then FSF2Info.Copyright := ReadStreamStr( aStream, aCount)
                  else if SameText( string( aFourCC), 'ICMT')     // ICMT sub chunk
                  then FSF2Info.Comment := ReadStreamStr( aStream, aCount)
                  else if SameText( string( aFourCC), 'ISFT')     // ISFT sub chunk
                  then FSF2Info.Creator := ReadStreamStr( aStream, aCount)
                  else aStream.Seek( aCount, soCurrent);          // Unknown sub chunk, that's fine, skip it
                end;
              end
              else if SameText( string( aFourCC), 'sdta')         // sdta list, the sample data
              then begin
                while aStream.Position < aStream.Size             // Read sub-chunks
                do begin
                  aStream.Read( aFourCC, SizeOf( aFourCC));
                  aStream.Read( aCount , SizeOf( aCount) );

                  if      SameText( string( aFourCC), 'smpl')     // smpl sub chunk - 16 bit data
                  then begin
                    if Length( aWaveData) = 0                     // Ignore multiple data sections
                    then begin
                      SetLength( aWaveData, 1);
                      SetLength( aWaveData[ 0], aCount div 2);

                      for j := 0 to aCount div 2 - 1
                      do begin
                        aStream.Read( IntSample, SizeOf( IntSample));
                        aWaveData[ 0, j] := 1.0 * IntSample / IntSample.MaxValue;

                        if   aWaveData[ 0, j] > FMaxValue
                        then FMaxValue := aWaveData[ 0, j];

                        if   aWaveData[ 0, j] < FMinValue
                        then FMinValue := aWaveData[ 0, j];
                      end;
                    end
                    else aStream.Seek( aCount, soCurrent);
                  end
                  else if SameText( string( aFourCC), 'sm24')     // sm24 sub chunk - 8 bit data to extend smpl data with
                  then begin
                    if    ( Length( aWaveData) = 1)               // Can only handle extra precission when we have dat and version >= 2.04
                    and   ((( FSF2Info.VersionMajor = 2) and ( FSF2Info.VersionMinor >= 4)) or ( FSF2Info.VersionMajor > 2))
                    then begin
                      aSize := Length( aWaveData[ 0]) div 2;

                      if   ( Odd( aSize) and ( aSize = aCount + 1))
                      or   ( aSize = aCount)
                      then begin
                        for j := 0 to aSize - 1
                        do begin
                          aStream.Read( ByteSample, SizeOf( ByteSample));
                          SampleValue := ( Int32( Round( Int16.MaxValue * aWaveData[ 0, j])) shl 8 + ByteSample) shl 8;
                          aWaveData[ 0, j] := SampleValue / SampleValue.MaxValue;

                          if   aWaveData[ 0, j] > FMaxValue
                          then FMaxValue := aWaveData[ 0, j];

                          if   aWaveData[ 0, j] < FMinValue
                          then FMinValue := aWaveData[ 0, j];
                        end;

                        if   Odd( aSize)
                        then aStream.Seek( 1, soCurrent);
                      end
                      else aStream.Seek( aCount, soCurrent);
                    end
                    else aStream.Seek( aCount, soCurrent);
                  end
                end;
              end
              else if SameText( string( aFourCC), 'pdta')         // pdta list
              then begin
                SetLength( FSF2PresetHeaders, 0);
                SetLength( FSF2PresetBags   , 0);
                SetLength( FSF2PModLists    , 0);
                SetLength( FSF2PGenLists    , 0);
                SetLength( FSF2Insts        , 0);
                SetLength( FSF2InstBags     , 0);
                SetLength( FSF2IModLists    , 0);
                SetLength( FSF2IGenLists    , 0);
                SetLength( FSF2Samples      , 0);

                while aStream.Position < aStream.Size             // Read sub-chunks
                do begin
                  aStream.Read( aFourCC, SizeOf( aFourCC));
                  aStream.Read( aCount , SizeOf( aCount) );

                  if      SameText( string( aFourCC), 'PHDR')     // PHDR sub chunk - head one of Hydra
                  then begin
                    if   ( aCount mod 38 = 0)
                    and  ( Length( FSF2PresetHeaders) = 0)
                    then begin
                      SetLength( FSF2PresetHeaders, Length( FSF2PresetHeaders) + Integer( aCount) div 38);
                      aStream.Read( FSF2PresetHeaders[ 0], aCount);
                    end
                    else aStream.Seek( aCount, soCurrent);
                  end
                  else if SameText( string( aFourCC), 'PBAG')     // PBAG sub chunk - head two of Hydra
                  then begin
                    if   ( aCount mod 4 = 0)
                    and  ( Length( FSF2PresetBags) = 0)
                    then begin
                      SetLength( FSF2PresetBags, Length( FSF2PresetBags) + Integer( aCount) div 4);
                      aStream.Read( FSF2PresetBags[ 0], aCount);
                    end
                    else aStream.Seek( aCount, soCurrent);
                  end
                  else if SameText( string( aFourCC), 'PMOD')     // PMOD sub chunk - head three of Hydra
                  then begin
                    if   ( aCount mod 10 = 0)
                    and  ( Length( FSF2PModLists) = 0)
                    then begin
                      SetLength( FSF2PModLists, Length( FSF2PModLists) + Integer( aCount) div 10);
                      aStream.Read( FSF2PModLists[ 0], aCount);
                    end
                    else aStream.Seek( aCount, soCurrent);
                  end
                  else if SameText( string( aFourCC), 'PGEN')     // PGEN sub chunk - head four of Hydra
                  then begin
                    if   ( aCount mod 4 = 0)
                    and  ( Length( FSF2PGenLists) = 0)
                    then begin
                      SetLength( FSF2PGenLists, Length( FSF2PGenLists) + Integer( aCount) div 4);
                      aStream.Read( FSF2PGenLists[ 0], aCount);
                    end
                    else aStream.Seek( aCount, soCurrent);
                  end
                  else if SameText( string( aFourCC), 'INST')     // INST sub chunk - head five of Hydra
                  then begin
                    if   ( aCount mod 22 = 0)
                    and  ( Length( FSF2Insts) = 0)
                    then begin
                      SetLength( FSF2Insts, Length( FSF2Insts) + Integer( aCount) div 22);
                      aStream.Read( FSF2Insts[ 0], aCount);
                    end
                    else aStream.Seek( aCount, soCurrent);
                  end
                  else if SameText( string( aFourCC), 'IBAG')     // IBAG sub chunk - head six of Hydra
                  then begin
                    if   ( aCount mod 4 = 0)
                    and  ( Length( FSF2InstBags) = 0)
                    then begin
                      SetLength( FSF2InstBags, Length( FSF2InstBags) + Integer( aCount) div 4);
                      aStream.Read( FSF2InstBags[ 0], aCount);
                    end
                    else aStream.Seek( aCount, soCurrent);
                  end
                  else if SameText( string( aFourCC), 'IMOD')     // IMOD sub chunk - head seven of Hydra
                  then begin
                    if   ( aCount mod 10 = 0)
                    and  ( Length( FSF2IModLists) = 0)
                    then begin
                      SetLength( FSF2IModLists, Length( FSF2IModLists) + Integer( aCount) div 10);
                      aStream.Read( FSF2IModLists[ 0], aCount);
                    end
                    else aStream.Seek( aCount, soCurrent);
                  end
                  else if SameText( string( aFourCC), 'IGEN')     // IGEN sub chunk - head eight of Hydra
                  then begin
                    if   ( aCount mod 4 = 0)
                    and  ( Length( FSF2IGenLists) = 0)
                    then begin
                      SetLength( FSF2IGenLists, Length( FSF2IGenLists) + Integer( aCount) div 4);
                      aStream.Read( FSF2IGenLists[ 0], aCount);
                    end
                    else aStream.Seek( aCount, soCurrent);
                  end
                  else if SameText( string( aFourCC), 'SHDR')     // SHDR sub chunk - head nine and last of Hydra
                  then begin
                    if   ( aCount mod 46 = 0)
                    and  ( Length( FSF2Samples) = 0)
                    then begin
                      SetLength( FSF2Samples, Length( FSF2Samples) + Integer( aCount) div 46);
                      aStream.Read( FSF2Samples[ 0], aCount);

                      FirstSampleRate := FSF2Samples[ 0].SampleRate;

                      for j := 0 to Length( FSF2Samples) - 1
                      do begin
                        if ( FSF2Samples[ j].SampleType and $8000) <> 0       // Any ROM samples?
                        then begin
                          AddError( 'ROM samples');
                          FSF2Info.Valid := False;                // We have no ROMs
                          Break;
                        end;
                      end;
                    end
                    else aStream.Seek( aCount, soCurrent);
                  end
                  else aStream.Seek( aCount, soCurrent);          // Unknown sub chunk, that's fine, skip it
                end; // while aStream.Position < aStream.Size // Read sub-chunks

                for j := 0 to Length( FSF2Samples) - 1
                do begin
                  {
                    TSF2Sample = packed record
                    public
                      SampleNameA     : array[ 1 .. 20] of AnsiChar;  //  0
                      StartPoint      : Cardinal;                     // 20
                      EndPoint        : Cardinal;                     // 24
                      StartLoop       : Cardinal;                     // 28
                      EndLoop         : Cardinal;                     // 32
                      SampleRate      : Cardinal;                     // 36
                      OriginalPitch   : Byte;                         // 40
                      PitchCorrection : Int8;                         // 41
                      SampleLink      : Word;                         // 42
                      SampleType      : Word;                         // 44
                    public                                            // 46 - ok
                      function  SampleName: string;
                    end;
                  }
                  if   not SameText( FSF2Samples[ j].SampleName, 'EOS')
                  then begin
                 // AddLabel( j, FSF2Samples[ j].StartPoint, FSF2Samples[ j].SampleName);
                    AddLoop ( j, FSF2Samples[ j].StartLoop, FSF2Samples[ j].EndLoop - FSF2Samples[ j].StartLoop, FSF2Samples[ j].SampleName);
                  end;
                end;
              end;
            end;
          end;
        end;

        if   FSF2Info.Valid
        then begin                                          // Export some format info
          FFormatInfo.WaveFormat   := fmtPCM;               // Word;     //  0 | Compression code ...
          FFormatInfo.ChannelCount := 1;                    // Word;     //  2 | Number of channels
          FFormatInfo.SampleRate   := FirstSampleRate;      // Cardinal; //  4 | Sample rate
          FFormatInfo.StreamSpeed  := FirstSampleRate * 4;  // Cardinal; //  8 | Average bytes per second
          FFormatInfo.BlockAlign   := 4;                    // Word;     // 12 | Block align
          FFormatInfo.BitCount     := 32;                   // Word;     // 14 | Bits per sample
          FFormatInfo.ExtraCount   := 0;                    // Word;     // 16 | Number of extra bytes

          if Length( aWaveData) > 0
          then begin
            FFrameCount := Length( aWaveData[ 0]);          // Cardinal, total amount of frames
            Result      := True;
          end
          else FFrameCount := 0;

        end
        else AddError( 'Unsupported sound font');
      end
      else AddErrorFmt( 'Unsupported file format - not a WAV, AIFF, AIFC or SF2 file? (seen: "%s")', [ string( aReader.FourCC)]);

      FIsValid := Result;
    end;


    procedure   TWaveFileReader.AddCueName( const anInfo: TLoopInfo; const aName: string); // overload;
    begin
      SetLength( FCueNames, Length( FCueNames) + 1);
      FCueNames[ Length( FCueNames) - 1].Id           := anInfo.Id;
      FCueNames[ Length( FCueNames) - 1].Name         := aName;
      FCueNames[ Length( FCueNames) - 1].CueType      := ctLoop;
      FCueNames[ Length( FCueNames) - 1].SampleLength := anInfo.LoopEnd - anInfo.LoopStart;
    end;


    procedure   TWaveFileReader.AddCueName( const anInfo: TCueInfo; aSampleLength: Cardinal; const aName: string; aType: TCueType); // overload;
    begin
      SetLength( FCueNames, Length( FCueNames) + 1);
      FCueNames[ Length( FCueNames) - 1].Id           := anInfo.Id;
      FCueNames[ Length( FCueNames) - 1].Name         := aName;
      FCueNames[ Length( FCueNames) - 1].CueType      := aType;
      FCueNames[ Length( FCueNames) - 1].SampleLength := aSampleLength;
    end;


    procedure   TWaveFileReader.AddCue( anId: Word; aPosition, aSampleLength: Cardinal; aType: TCueType; const aName: string);
    begin
      SetLength( FCueInfo, Length( FCueInfo ) + 1);

      FCueInfo[ Length( FCueInfo ) - 1].Id           := anId;
      FCueInfo[ Length( FCueInfo ) - 1].Position     := aPosition;
      FCueInfo[ Length( FCueInfo ) - 1].Chunk        := 'data';
      FCueInfo[ Length( FCueInfo ) - 1].ChunkStart   := 0;
      FCueInfo[ Length( FCueInfo ) - 1].BlockStart   := 0;
      FCueInfo[ Length( FCueInfo ) - 1].SampleOffset := aPosition;
      AddCueName( FCueInfo[ Length( FCueInfo) - 1], aSampleLEngth, aName, aType);
    end;


    procedure   TWaveFileReader.AddLabel( anId: Word; aPosition: Cardinal; const aName: string);
    begin
      AddCue( anId, aPosition, 0, ctLabel, aName);
    end;


    procedure   TWaveFileReader.AddRegion( anId: Word; aPosition, aSampleLength: Cardinal; const aName: string);
    begin
      AddCue( anId, aPosition, aSampleLength, ctRegion, aName);
    end;


    procedure   TWaveFileReader.AddLoop( anId: Word; aPosition, aSampleLEngth: Cardinal; const aName: string);
    var
      lend : Cardinal;
    begin
      SetLength( FLoopInfo, Length( FLoopInfo) + 1);
      lend := aPosition + aSampleLEngth;

      FLoopInfo[ Length( FLoopInfo) - 1].Id         := anId;      // Identifier : Cardinal; // Loop ID - may be mapped to an ASCII name trough an "cue " chunk
      FLoopInfo[ Length( FLoopInfo) - 1].LoopType   := 0;         // LoopType   : Cardinal; // 0 - forward, 1 - alternating fwd bwd, 2 - backward, 3-31: reserved, 32.. sampler specific
      FLoopInfo[ Length( FLoopInfo) - 1].LoopStart  := aPosition; // LoopStart  : Cardinal; // First sample played for the loop in "data" chunk
      FLoopInfo[ Length( FLoopInfo) - 1].LoopEnd    := lend;      // LoopEnd    : Cardinal; // Last  sample played for the loop in "data" chunk
      FLoopInfo[ Length( FLoopInfo) - 1].Fraction   := 0;         // Fraction   : Cardinal; // fine tuning for loop fractional areas .. $80000000 -> 1 / 2 sample
      FLoopInfo[ Length( FLoopInfo) - 1].PlayCount  := 1;         // PlayCount  : Cardinal; // Repeat count for the loop
      AddCueName( FLoopInfo[ Length( FLoopInfo) - 1], aName);
    end;


    procedure   TWaveFileReader.AddError( const aMsg: string);
    begin
      if FErrorStr = ''
      then FErrorStr :=                    aMsg
      else FErrorStr := FErrorStr + '; ' + aMsg;
    end;


    procedure   TWaveFileReader.AddErrorFmt( const aFmt: string; const anArgs: array of const);
    begin
      AddError( Format( aFmt, anArgs, AppLocale));
    end;


//  public

    function    TWaveFileReader.ReadData( const aReader: TRiffReaderWriter; var aWaveData: TWaveData): Boolean; // overload;
    var
      i : Integer;
    begin
      FIsValid  := False;
      Result    := False;
      FErrorStr := '';

      for i := 0 to Length( aWaveData) - 1
      do SetLength( aWaveData[ i], 0);

      SetLength( aWaveData, 0);

      try
        Result := ParseData( aReader, aWaveData);

        if   AutoNormalize
        then NormalizeData( aWaveData);
      except
        on E: Exception           // Ignore exceptions .. IsValid will signal validity
        do AddError( E.ToString);
      end;
    end;


    procedure   TWaveFileReader.NormalizeData( const aWaveData: TWaveData); // overload;
    var
      aScale : TSignal;
      i      : Integer;
      j      : Integer;
    begin
      aScale := Max( Abs( MinValue), Abs( MaxValue));

      if   aScale <> 0
      then begin
        aScale := 1.0 / aScale;

        FMinValue :=  1.0;
        FMaxValue := -1.0;

        for i := 0 to Length( aWaveData) - 1
        do begin
          for j := 0 to Length( aWaveData[ i]) - 1
          do begin
            aWaveData[ i, j] := aScale * aWaveData[ i, j];

            if   aWaveData[ i, j] < FMinValue
            then FMinValue := aWaveData[ i, j];

            if   aWaveData[ i, j] > FMaxValue
            then FMaxValue := aWaveData[ i, j];
          end;
        end;
      end;
    end;


    function    TWaveFileReader.HasLoops: Boolean;
    begin
      Result := LoopInfoCount > 0;
    end;


    function    TWaveFileReader.Dump: string;
    const
      CRLF: string = #10;
    var
      i        : Integer;
      aCueName : TCueName;
      anId     : Cardinal;
    begin
      Result := ''
        + Format( 'IsValid                     : %s' + CRLF, [ BoolToStr( IsValid      , True)], AppLocale)
        + Format( 'StreamFormat                : %u' + CRLF, [ StreamFormat                   ], AppLocale)
        + Format( 'ChannelCount                : %u' + CRLF, [ ChannelCount                   ], AppLocale)
        + Format( 'SampleRate                  : %u' + CRLF, [ SampleRate                     ], AppLocale)
        + Format( 'Duration                    : %g' + CRLF, [ Duration                       ], AppLocale)
        + Format( 'FrameCount                  : %u' + CRLF, [ FrameCount                     ], AppLocale)
        + Format( 'MinValue                    : %g' + CRLF, [ MinValue                       ], AppLocale)
        + Format( 'MaxValue                    : %g' + CRLF, [ MaxValue                       ], AppLocale)
        + Format( 'HasAcid                     : %s' + CRLF, [ BoolToStr( HasAcid      , True)], AppLocale)
        + Format( 'HasSampleInfo               : %s' + CRLF, [ BoolToStr( HasSampleInfo, True)], AppLocale)
        + Format( 'ErrorStr                    : %s' + CRLF, [ ErrorStr                       ], AppLocale)
        + Format( 'BlockAlign                  : %u' + CRLF, [ BlockAlign                     ], AppLocale)
        + CRLF
        + Format( 'FormatInfo.WaveFormat       : %u' + CRLF, [ FormatInfo.WaveFormat          ], AppLocale)
        + Format( 'FormatInfo.ChannelCount     : %u' + CRLF, [ FormatInfo.ChannelCount        ], AppLocale)
        + Format( 'FormatInfo.SampleRate       : %u' + CRLF, [ FormatInfo.SampleRate          ], AppLocale)
        + Format( 'FormatInfo.StreamSpeed      : %u' + CRLF, [ FormatInfo.StreamSpeed         ], AppLocale)
        + Format( 'FormatInfo.BlockAlign       : %u' + CRLF, [ FormatInfo.BlockAlign          ], AppLocale)
        + Format( 'FormatInfo.BitCount         : %u' + CRLF, [ FormatInfo.BitCount            ], AppLocale)
        + Format( 'FormatInfo.ExtraCount       : %u' + CRLF, [ FormatInfo.ExtraCount          ], AppLocale)
        + CRLF
        ;

      if   HasAcid
      then begin
        Result := Result
        + Format( 'AcidInfo.FileType           : %u' + CRLF, [ AcidInfo.FileType              ], AppLocale)
        + Format( 'AcidInfo.RootNote           : %u' + CRLF, [ AcidInfo.RootNote              ], AppLocale)
        + Format( 'AcidInfo.UnSpec_1           : %u' + CRLF, [ AcidInfo.UnSpec_1              ], AppLocale)
        + Format( 'AcidInfo.UnSpec_2           : %g' + CRLF, [ AcidInfo.UnSpec_2              ], AppLocale)
        + Format( 'AcidInfo.NumBeats           : %u' + CRLF, [ AcidInfo.NumBeats              ], AppLocale)
        + Format( 'AcidInfo.MeterNumer         : %u' + CRLF, [ AcidInfo.MeterNumer            ], AppLocale)
        + Format( 'AcidInfo.MeterDenom         : %u' + CRLF, [ AcidInfo.MeterDenom            ], AppLocale)
        + Format( 'AcidInfo.Tempo              : %g' + CRLF, [ AcidInfo.Tempo                 ], AppLocale)
        + CRLF
        ;
      end;

      if   HasSampleInfo
      then begin
        Result := Result
        + Format( 'SampleInfo.Manufacturer     : %u' + CRLF, [ SampleInfo.Manufacturer        ], AppLocale)
        + Format( 'SampleInfo.Product          : %u' + CRLF, [ SampleInfo.Product             ], AppLocale)
        + Format( 'SampleInfo.SamplePeriod     : %u' + CRLF, [ SampleInfo.SamplePeriod        ], AppLocale)
        + Format( 'SampleInfo.MIDIUnityNote    : %u' + CRLF, [ SampleInfo.MIDIUnityNote       ], AppLocale)
        + Format( 'SampleInfo.MIDIPitchFraction: %u' + CRLF, [ SampleInfo.MIDIPitchFraction   ], AppLocale)
        + Format( 'SampleInfo.SMPTEFormat      : %u' + CRLF, [ SampleInfo.SMPTEFormat         ], AppLocale)
        + Format( 'SampleInfo.SMPTEOffset      : %u' + CRLF, [ SampleInfo.SMPTEOffset         ], AppLocale)
        + Format( 'SampleInfo.SampleLoops      : %u' + CRLF, [ SampleInfo.SampleLoops         ], AppLocale)
        + Format( 'SampleInfo.SamplerDataSize  : %u' + CRLF, [ SampleInfo.SamplerDataSize     ], AppLocale)
        + CRLF
        ;
      end;

      Result := Result
        + Format( 'LoopInfoCount               : %d' + CRLF, [ LoopInfoCount                  ], AppLocale)
        + Format( 'CueInfoCount                : %d' + CRLF, [ CueInfoCount                   ], AppLocale)
        + Format( 'CueNameCount                : %d' + CRLF, [ CueNameCount                   ], AppLocale)
        + CRLF
        + CRLF
        + '---- Loop and Cue info:'
        + CRLF
        + CRLF
        ;

      for i := 0 to LoopInfoCount - 1
      do begin
        anId := LoopInfo[ i].Id;
        Result := Result
        + Format( 'LoopInfo[ %d].Identifier    : %u' + CRLF, [ i, LoopInfo[ i  ].Id           ], AppLocale)
        + Format( 'LoopInfo[ %d].LoopType      : %u' + CRLF, [ i, LoopInfo[ i  ].LoopType     ], AppLocale)
        + Format( 'LoopInfo[ %d].LoopStart     : %u' + CRLF, [ i, LoopInfo[ i  ].LoopStart    ], AppLocale)
        + Format( 'LoopInfo[ %d].LoopEnd       : %u' + CRLF, [ i, LoopInfo[ i  ].LoopEnd      ], AppLocale)
        + Format( 'LoopInfo[ %d].Fraction      : %u' + CRLF, [ i, LoopInfo[ i  ].Fraction     ], AppLocale)
        + Format( 'LoopInfo[ %d].PlayCount     : %u' + CRLF, [ i, LoopInfo[ i  ].PlayCount    ], AppLocale)
        + CRLF
        + Format( 'CueName[ %d].Id             : %u' + CRLF, [ i, CueName[ anId].Id           ], AppLocale)
        + Format( 'CueName[ %d].Name           : %s' + CRLF, [ i, CueName[ anId].Name         ], AppLocale)
        + Format( 'CueName[ %d].CueType        : %s' + CRLF, [ i, CueName[ anId].CueTypeName  ], AppLocale)
        + Format( 'CueName[ %d].SampleLength   : %u' + CRLF, [ i, CueName[ anId].SampleLength ], AppLocale)
        + CRLF
        ;
      end;

      for i := 0 to CueInfoCount - 1
      do begin
        anId     := CueInfo[ i].Id;
        aCueName := CueName[ i];

        Result := Result
        + Format( 'CueInfo[ %d].Id             : %u' + CRLF, [ i, CueInfo[ i   ].Id          ], AppLocale)
        + Format( 'CueInfo[ %d].Position       : %u' + CRLF, [ i, CueInfo[ i   ].Position    ], AppLocale)
        + Format( 'CueInfo[ %d].Chunk          : %s' + CRLF, [ i, string( CueInfo[ i].Chunk) ], AppLocale)
        + Format( 'CueInfo[ %d].ChunkStart     : %u' + CRLF, [ i, CueInfo[ i   ].ChunkStart  ], AppLocale)
        + Format( 'CueInfo[ %d].BlockStart     : %u' + CRLF, [ i, CueInfo[ i   ].BlockStart  ], AppLocale)
        + Format( 'CueInfo[ %d].SampleOffset   : %u' + CRLF, [ i, CueInfo[ i   ].SampleOffset], AppLocale)
        + CRLF
        + Format( 'CueName[ %d].Id             : %u' + CRLF, [ i, CueName[ anId].Id          ], AppLocale)
        + Format( 'CueName[ %d].Name           : %s' + CRLF, [ i, CueName[ anId].Name        ], AppLocale)
        + Format( 'CueName[ %d].CueType        : %s' + CRLF, [ i, CueName[ anId].CueTypeName ], AppLocale)
        + Format( 'CueName[ %d].SampleLength   : %u' + CRLF, [ i, CueName[ anId].SampleLength ], AppLocale)
        + CRLF
        ;
      end;

      // + Format( ': %d' + CRLF, [ ], AppLocale)
      // Data[ anIndex: Integer] : TSignal
    end;



end.

